#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Mapping, Optional

import pytest
from airbyte_cdk.sources.file_based.exceptions import ConfigValidationError, SchemaInferenceError
from airbyte_cdk.sources.file_based.schema_helpers import (
    ComparableType,
    SchemaType,
    conforms_to_schema,
    merge_schemas,
    type_mapping_to_jsonschema,
)

COMPLETE_CONFORMING_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}


NONCONFORMING_EXTRA_COLUMN_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
    "column_x": "extra",
}

CONFORMING_WITH_MISSING_COLUMN_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
}

CONFORMING_WITH_NARROWER_TYPE_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": True,
    "number_field": True,
    "string_field": True,
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

NONCONFORMING_WIDER_TYPE_RECORD = {
    "null_field": "not None",
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

NONCONFORMING_NON_OBJECT_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": "not an object",
}

NONCONFORMING_NON_ARRAY_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": "not an array",
    "object_field": {"col": "val"},
}

CONFORMING_MIXED_TYPE_NARROWER_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

NONCONFORMING_MIXED_TYPE_WIDER_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

CONFORMING_MIXED_TYPE_WITHIN_TYPE_RANGE_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "val1",
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

NONCONFORMING_INVALID_ARRAY_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": ["this should not be an array"],
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}

NONCONFORMING_TOO_WIDE_ARRAY_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "okay",
    "array_field": ["val1", "val2"],
    "object_field": {"col": "val"},
}


CONFORMING_NARROWER_ARRAY_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": "okay",
    "array_field": [1, 2],
    "object_field": {"col": "val"},
}


NONCONFORMING_INVALID_OBJECT_RECORD = {
    "null_field": None,
    "boolean_field": True,
    "integer_field": 1,
    "number_field": 1.5,
    "string_field": {"this": "should not be an object"},
    "array_field": [1.1, 2.2],
    "object_field": {"col": "val"},
}


SCHEMA = {
    "type": "object",
    "properties": {
        "null_field": {"type": "null"},
        "boolean_field": {"type": "boolean"},
        "integer_field": {"type": "integer"},
        "number_field": {"type": "number"},
        "string_field": {"type": "string"},
        "array_field": {
            "type": "array",
            "items": {
                "type": "number",
            },
        },
        "object_field": {"type": "object"},
    },
}


@pytest.mark.parametrize(
    "record,schema,expected_result",
    [
        pytest.param(COMPLETE_CONFORMING_RECORD, SCHEMA, True, id="record-conforms"),
        pytest.param(NONCONFORMING_EXTRA_COLUMN_RECORD, SCHEMA, False, id="nonconforming-extra-column"),
        pytest.param(CONFORMING_WITH_MISSING_COLUMN_RECORD, SCHEMA, True, id="record-conforms-with-missing-column"),
        pytest.param(CONFORMING_WITH_NARROWER_TYPE_RECORD, SCHEMA, True, id="record-conforms-with-narrower-type"),
        pytest.param(NONCONFORMING_WIDER_TYPE_RECORD, SCHEMA, False, id="nonconforming-wider-type"),
        pytest.param(NONCONFORMING_NON_OBJECT_RECORD, SCHEMA, False, id="nonconforming-string-is-not-an-object"),
        pytest.param(NONCONFORMING_NON_ARRAY_RECORD, SCHEMA, False, id="nonconforming-string-is-not-an-array"),
        pytest.param(NONCONFORMING_TOO_WIDE_ARRAY_RECORD, SCHEMA, False, id="nonconforming-array-values-too-wide"),
        pytest.param(CONFORMING_NARROWER_ARRAY_RECORD, SCHEMA, True, id="conforming-array-values-narrower-than-schema"),
        pytest.param(NONCONFORMING_INVALID_ARRAY_RECORD, SCHEMA, False, id="nonconforming-array-is-not-a-string"),
        pytest.param(NONCONFORMING_INVALID_OBJECT_RECORD, SCHEMA, False, id="nonconforming-object-is-not-a-string"),
    ],
)
def test_conforms_to_schema(record: Mapping[str, Any], schema: Mapping[str, Any], expected_result: bool) -> None:
    assert conforms_to_schema(record, schema) == expected_result


def test_comparable_types() -> None:
    assert ComparableType.OBJECT > ComparableType.STRING
    assert ComparableType.STRING > ComparableType.NUMBER
    assert ComparableType.NUMBER > ComparableType.INTEGER
    assert ComparableType.INTEGER > ComparableType.BOOLEAN
    assert ComparableType["OBJECT"] == ComparableType.OBJECT


@pytest.mark.parametrize(
    "schema1,schema2,expected_result",
    [
        pytest.param({}, {}, {}, id="empty-schemas"),
        pytest.param({"a": None}, {}, None, id="null-value-in-schema"),
        pytest.param({"a": {"type": "integer"}}, {}, {"a": {"type": "integer"}}, id="single-key-schema1"),
        pytest.param({}, {"a": {"type": "integer"}}, {"a": {"type": "integer"}}, id="single-key-schema2"),
        pytest.param({"a": {"type": "integer"}}, {"a": {"type": "integer"}}, {"a": {"type": "integer"}}, id="single-key-both-schemas"),
        pytest.param({"a": {"type": "integer"}}, {"a": {"type": "number"}}, {"a": {"type": "number"}}, id="single-key-schema2-is-wider"),
        pytest.param({"a": {"type": "number"}}, {"a": {"type": "integer"}}, {"a": {"type": "number"}}, id="single-key-schema1-is-wider"),
        pytest.param({"a": {"type": "array"}}, {"a": {"type": "integer"}}, None, id="single-key-with-array-schema1"),
        pytest.param({"a": {"type": "integer"}}, {"a": {"type": "array"}}, None, id="single-key-with-array-schema2"),
        pytest.param(
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            id="single-key-same-object",
        ),
        pytest.param(
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            {"a": {"type": "object", "properties": {"b": {"type": "string"}}}},
            None,
            id="single-key-different-objects",
        ),
        pytest.param(
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            {"a": {"type": "number"}},
            None,
            id="single-key-with-object-schema1",
        ),
        pytest.param(
            {"a": {"type": "number"}},
            {"a": {"type": "object", "properties": {"b": {"type": "integer"}}}},
            None,
            id="single-key-with-object-schema2",
        ),
        pytest.param(
            {"a": {"type": "array", "items": {"type": "number"}}},
            {"a": {"type": "array", "items": {"type": "number"}}},
            {"a": {"type": "array", "items": {"type": "number"}}},
            id="equal-arrays-in-both-schemas",
        ),
        pytest.param(
            {"a": {"type": "array", "items": {"type": "integer"}}},
            {"a": {"type": "array", "items": {"type": "number"}}},
            None,
            id="different-arrays-in-both-schemas",
        ),
        pytest.param(
            {"a": {"type": "integer"}, "b": {"type": "string"}},
            {"c": {"type": "number"}},
            {"a": {"type": "integer"}, "b": {"type": "string"}, "c": {"type": "number"}},
            id="",
        ),
        pytest.param({"a": {"type": "invalid_type"}}, {"b": {"type": "integer"}}, None, id="invalid-type"),
    ],
)
def test_merge_schemas(schema1: SchemaType, schema2: SchemaType, expected_result: Optional[SchemaType]) -> None:
    if expected_result is not None:
        assert merge_schemas(schema1, schema2) == expected_result
    else:
        with pytest.raises(SchemaInferenceError):
            merge_schemas(schema1, schema2)


@pytest.mark.parametrize(
    "type_mapping,expected_schema,expected_exc_msg",
    [
        pytest.param(
            '{"col1": "null", "col2": "array", "col3": "boolean", "col4": "float", "col5": "integer", "col6": "number", "col7": "object", "col8": "string"}',
            {
                "type": "object",
                "properties": {
                    "col1": {"type": "null"},
                    "col2": {"type": "array"},
                    "col3": {"type": "boolean"},
                    "col4": {"type": "number"},
                    "col5": {"type": "integer"},
                    "col6": {"type": "number"},
                    "col7": {"type": "object"},
                    "col8": {"type": "string"},
                },
            },
            None,
            id="valid_all_types",
        ),
        pytest.param(
            '{"col1 ": " string", "col2":  " integer"}',
            {"type": "object", "properties": {"col1": {"type": "string"}, "col2": {"type": "integer"}}},
            None,
            id="valid_extra_spaces",
        ),
        pytest.param(
            "",
            None,
            None,
            id="valid_empty_string",
        ),
        pytest.param(
            '{"col1": "x", "col2": "integer"}',
            None,
            "Invalid type 'x' for property 'col1'",
            id="invalid_type",
        ),
        pytest.param(
            '{"col1": "", "col2": "integer"}',
            None,
            "Invalid input schema",
            id="invalid_missing_type",
        ),
        pytest.param(
            '{"": "string", "col2": "integer"}',
            None,
            "Invalid input schema",
            id="invalid_missing_name",
        ),
        pytest.param(
            '{"type": "object", "properties": {"col1": {"type": "string"}, "col2": {"type": "integer"}}}',
            None,
            "Invalid input schema; nested schemas are not supported.",
            id="invalid_nested_input_string",
        ),
        pytest.param(
            '{"type": "object", "properties": {"col1": {"type": "string"}, "col2": {"type": "integer"}}}',
            None,
            "Invalid input schema; nested schemas are not supported.",
            id="invalid_nested_input_json",
        ),
    ],
)
def test_type_mapping_to_jsonschema(
    type_mapping: Mapping[str, Any], expected_schema: Optional[Mapping[str, Any]], expected_exc_msg: Optional[str]
) -> None:
    if expected_exc_msg:
        with pytest.raises(ConfigValidationError) as exc:
            type_mapping_to_jsonschema(type_mapping)
        assert expected_exc_msg in exc.value.args[0]
    else:
        assert type_mapping_to_jsonschema(type_mapping) == expected_schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import inspect
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generator, List, Mapping, Tuple

from airbyte_cdk.models import AirbyteStream, ConfiguredAirbyteCatalog, ConfiguredAirbyteStream, SyncMode
from airbyte_cdk.sources.utils.schema_helpers import ResourceSchemaLoader


def package_name_from_class(cls: object) -> str:
    """Find the package name given a class name"""
    module: Any = inspect.getmodule(cls)
    return module.__name__.split(".")[0]


class StreamStateMixin:
    def get_stream_state(self, name: str) -> Any:
        """Get state of stream with corresponding name"""
        raise NotImplementedError

    def set_stream_state(self, name: str, state: Any):
        """Set state of stream with corresponding name"""
        raise NotImplementedError

    def stream_has_state(self, name: str) -> bool:
        """Tell if stream supports incremental sync"""
        return False


class BaseClient(StreamStateMixin, ABC):
    """Base client for API"""

    schema_loader_class = ResourceSchemaLoader

    def __init__(self, **kwargs):
        package_name = package_name_from_class(self.__class__)
        self._schema_loader = self.schema_loader_class(package_name)
        self._stream_methods = self._enumerate_methods()

    def _enumerate_methods(self) -> Mapping[str, Callable]:
        """Detect available streams and return mapping"""
        prefix = "stream__"
        mapping = {}
        methods = inspect.getmembers(self.__class__, predicate=inspect.isfunction)
        for name, method in methods:
            if name.startswith(prefix):
                mapping[name[len(prefix) :]] = getattr(self, name)

        return mapping

    @staticmethod
    def _get_fields_from_stream(stream: AirbyteStream) -> List[str]:
        return list(stream.json_schema.get("properties", {}).keys())

    def _get_stream_method(self, name: str) -> Callable:
        method = self._stream_methods.get(name)
        if not method:
            raise ValueError(f"Client does not know how to read stream `{name}`")
        return method

    def read_stream(self, stream: AirbyteStream) -> Generator[Dict[str, Any], None, None]:
        """Yield records from stream"""
        method = self._get_stream_method(stream.name)
        fields = self._get_fields_from_stream(stream)

        for message in method(fields=fields):
            yield dict(message)

    @property
    def streams(self) -> Generator[AirbyteStream, None, None]:
        """List of available streams"""
        for name, method in self._stream_methods.items():
            supported_sync_modes = [SyncMode.full_refresh]
            source_defined_cursor = False
            if self.stream_has_state(name):
                supported_sync_modes += [SyncMode.incremental]
                source_defined_cursor = True

            yield AirbyteStream(
                name=name,
                json_schema=self._schema_loader.get_schema(name),
                supported_sync_modes=supported_sync_modes,
                source_defined_cursor=source_defined_cursor,
            )

    @abstractmethod
    def health_check(self) -> Tuple[bool, str]:
        """Check if service is up and running"""


def configured_catalog_from_client(client: BaseClient) -> ConfiguredAirbyteCatalog:
    """Helper to generate configured catalog for testing"""
    catalog = ConfiguredAirbyteCatalog(streams=[ConfiguredAirbyteStream(stream=stream) for stream in client.streams])

    return catalog



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from typing import Any, Dict, List

from airbyte_protocol.models import ConfiguredAirbyteCatalog, SyncMode


class CatalogBuilder:
    def __init__(self) -> None:
        self._streams: List[Dict[str, Any]] = []

    def with_stream(self, name: str, sync_mode: SyncMode) -> "CatalogBuilder":
        self._streams.append(
            {
                "stream": {
                    "name": name,
                    "json_schema": {},
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "source_defined_primary_key": [["id"]],
                },
                "primary_key": [["id"]],
                "sync_mode": sync_mode.name,
                "destination_sync_mode": "overwrite",
            }
        )
        return self

    def build(self) -> ConfiguredAirbyteCatalog:
        return ConfiguredAirbyteCatalog.parse_obj({"streams": self._streams})



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import logging
from abc import ABC
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union

import requests
from airbyte_cdk.models import (
    AirbyteStream,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    ConnectorSpecification,
    DestinationSyncMode,
    SyncMode,
)
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.http import HttpStream
from airbyte_cdk.sources.streams.http.requests_native_auth import Oauth2Authenticator
from requests.auth import AuthBase


class SourceTestFixture(AbstractSource):
    """
    This is a concrete implementation of a Source connector that provides implementations of all the methods needed to run sync
    operations. For simplicity, it also overrides functions that read from files in favor of returning the data directly avoiding
    the need to load static files (ex. spec.yaml, config.json, configured_catalog.json) into the unit-test package.
    """

    def __init__(self, streams: Optional[List[Stream]] = None, authenticator: Optional[AuthBase] = None):
        self._streams = streams
        self._authenticator = authenticator

    def spec(self, logger: logging.Logger) -> ConnectorSpecification:
        return ConnectorSpecification(
            connectionSpecification={
                "$schema": "http://json-schema.org/draft-07/schema#",
                "title": "Test Fixture Spec",
                "type": "object",
                "required": ["api_token"],
                "properties": {
                    "api_token": {
                        "type": "string",
                        "title": "API token",
                        "description": "The token used to authenticate requests to the API.",
                        "airbyte_secret": True,
                    }
                },
            }
        )

    def read_config(self, config_path: str) -> Mapping[str, Any]:
        return {"api_token": "just_some_token"}

    @classmethod
    def read_catalog(cls, catalog_path: str) -> ConfiguredAirbyteCatalog:
        return ConfiguredAirbyteCatalog(
            streams=[
                ConfiguredAirbyteStream(
                    stream=AirbyteStream(
                        name="http_test_stream",
                        json_schema={},
                        supported_sync_modes=[SyncMode.full_refresh, SyncMode.incremental],
                        default_cursor_field=["updated_at"],
                        source_defined_cursor=True,
                        source_defined_primary_key=[["id"]],
                    ),
                    sync_mode=SyncMode.full_refresh,
                    destination_sync_mode=DestinationSyncMode.overwrite,
                )
            ]
        )

    def check_connection(self, *args, **kwargs) -> Tuple[bool, Optional[Any]]:
        return True, ""

    def streams(self, *args, **kwargs) -> List[Stream]:
        return [HttpTestStream(authenticator=self._authenticator)]


class HttpTestStream(HttpStream, ABC):
    url_base = "https://airbyte.com/api/v1/"

    def supports_incremental(self):
        return True

    @property
    def availability_strategy(self):
        return None

    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        return "id"

    def path(
        self,
        *,
        stream_state: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> str:
        return "cast"

    def parse_response(
        self,
        response: requests.Response,
        *,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> Iterable[Mapping]:
        body = response.json() or {}
        return body["records"]

    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
        return None

    def get_json_schema(self) -> Mapping[str, Any]:
        return {}


def fixture_mock_send(self, request, **kwargs) -> requests.Response:
    """
    Helper method that can be used by a test to patch the Session.send() function and mock the outbound send operation to provide
    faster and more reliable responses compared to actual API requests
    """
    response = requests.Response()
    response.request = request
    response.status_code = 200
    response.headers = {"header": "value"}
    response_body = {
        "records": [
            {"id": 1, "name": "Celine Song", "position": "director"},
            {"id": 2, "name": "Shabier Kirchner", "position": "cinematographer"},
            {"id": 3, "name": "Christopher Bear", "position": "composer"},
            {"id": 4, "name": "Daniel Rossen", "position": "composer"},
        ]
    }
    response._content = json.dumps(response_body).encode("utf-8")
    return response


class SourceFixtureOauthAuthenticator(Oauth2Authenticator):
    """
    Test OAuth authenticator that only overrides the request and response aspect of the authenticator flow
    """

    def refresh_access_token(self) -> Tuple[str, int]:
        response = requests.request(method="POST", url=self.get_token_refresh_endpoint(), params={})
        response.raise_for_status()
        return "some_access_token", 1800  # Mock oauth response values to be used during the data retrieval step



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import os

CLOUD_DEPLOYMENT_MODE = "cloud"


def is_cloud_environment() -> bool:
    """
    Returns True if the connector is running in a cloud environment, False otherwise.

    The function checks the value of the DEPLOYMENT_MODE environment variable which is set by the platform.
    This function can be used to determine whether stricter security measures should be applied.
    """
    deployment_mode = os.environ.get("DEPLOYMENT_MODE", "")
    return deployment_mode.casefold() == CLOUD_DEPLOYMENT_MODE



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""
The AirbyteEntrypoint is important because it is a service layer that orchestrate how we execute commands from the
[common interface](https://docs.airbyte.com/understanding-airbyte/airbyte-protocol#common-interface) through the source Python
implementation. There is some logic about which message we send to the platform and when which is relevant for integration testing. Other
than that, there are integrations point that are annoying to integrate with using Python code:
* Sources communicate with the platform using stdout. The implication is that the source could just print every message instead of
    returning things to source.<method> or to using the message repository. WARNING: As part of integration testing, we will not support
    messages that are simply printed. The reason is that capturing stdout relies on overriding sys.stdout (see
    https://docs.python.org/3/library/contextlib.html#contextlib.redirect_stdout) which clashes with how pytest captures logs and brings
    considerations for multithreaded applications. If code you work with uses `print` statements, please migrate to
    source.message_repository to emit those messages
* The entrypoint interface relies on file being written on the file system
"""

import json
import logging
import tempfile
import traceback
from io import StringIO
from pathlib import Path
from typing import Any, List, Mapping, Optional, Union

from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.exception_handler import assemble_uncaught_exception
from airbyte_cdk.logger import AirbyteLogFormatter
from airbyte_cdk.sources import Source
from airbyte_protocol.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteStateMessage,
    AirbyteStreamStatus,
    ConfiguredAirbyteCatalog,
    Level,
    TraceType,
    Type,
)
from pydantic.error_wrappers import ValidationError


class EntrypointOutput:
    def __init__(self, messages: List[str], uncaught_exception: Optional[BaseException] = None):
        try:
            self._messages = [self._parse_message(message) for message in messages]
        except ValidationError as exception:
            raise ValueError("All messages are expected to be AirbyteMessage") from exception

        if uncaught_exception:
            self._messages.append(assemble_uncaught_exception(type(uncaught_exception), uncaught_exception).as_airbyte_message())

    @staticmethod
    def _parse_message(message: str) -> AirbyteMessage:
        try:
            return AirbyteMessage.parse_obj(json.loads(message))
        except (json.JSONDecodeError, ValidationError):
            # The platform assumes that logs that are not of AirbyteMessage format are log messages
            return AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message=message))

    @property
    def records_and_state_messages(self) -> List[AirbyteMessage]:
        return self._get_message_by_types([Type.RECORD, Type.STATE])

    @property
    def records(self) -> List[AirbyteMessage]:
        return self._get_message_by_types([Type.RECORD])

    @property
    def state_messages(self) -> List[AirbyteMessage]:
        return self._get_message_by_types([Type.STATE])

    @property
    def most_recent_state(self) -> Any:
        state_messages = self._get_message_by_types([Type.STATE])
        if not state_messages:
            raise ValueError("Can't provide most recent state as there are no state messages")
        return state_messages[-1].state.data

    @property
    def logs(self) -> List[AirbyteMessage]:
        return self._get_message_by_types([Type.LOG])

    @property
    def trace_messages(self) -> List[AirbyteMessage]:
        return self._get_message_by_types([Type.TRACE])

    @property
    def analytics_messages(self) -> List[AirbyteMessage]:
        return self._get_trace_message_by_trace_type(TraceType.ANALYTICS)

    @property
    def errors(self) -> List[AirbyteMessage]:
        return self._get_trace_message_by_trace_type(TraceType.ERROR)

    def get_stream_statuses(self, stream_name: str) -> List[AirbyteStreamStatus]:
        status_messages = map(
            lambda message: message.trace.stream_status.status,
            filter(
                lambda message: message.trace.stream_status.stream_descriptor.name == stream_name,
                self._get_trace_message_by_trace_type(TraceType.STREAM_STATUS),
            ),
        )
        return list(status_messages)

    def _get_message_by_types(self, message_types: List[Type]) -> List[AirbyteMessage]:
        return [message for message in self._messages if message.type in message_types]

    def _get_trace_message_by_trace_type(self, trace_type: TraceType) -> List[AirbyteMessage]:
        return [message for message in self._get_message_by_types([Type.TRACE]) if message.trace.type == trace_type]


def read(
    source: Source,
    config: Mapping[str, Any],
    catalog: ConfiguredAirbyteCatalog,
    state: Optional[List[AirbyteStateMessage]] = None,
    expecting_exception: bool = False,
) -> EntrypointOutput:
    """
    config and state must be json serializable

    :param expecting_exception: By default if there is an uncaught exception, the exception will be printed out. If this is expected, please
        provide expecting_exception=True so that the test output logs are cleaner
    """
    log_capture_buffer = StringIO()
    stream_handler = logging.StreamHandler(log_capture_buffer)
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(AirbyteLogFormatter())
    parent_logger = logging.getLogger("")
    parent_logger.addHandler(stream_handler)

    with tempfile.TemporaryDirectory() as tmp_directory:
        tmp_directory_path = Path(tmp_directory)
        args = [
            "read",
            "--config",
            make_file(tmp_directory_path / "config.json", config),
            "--catalog",
            make_file(tmp_directory_path / "catalog.json", catalog.json()),
        ]
        if state is not None:
            args.extend(
                [
                    "--state",
                    make_file(tmp_directory_path / "state.json", f"[{','.join([stream_state.json() for stream_state in state])}]"),
                ]
            )
        args.append("--debug")
        source_entrypoint = AirbyteEntrypoint(source)
        parsed_args = source_entrypoint.parse_args(args)

        messages = []
        uncaught_exception = None
        try:
            for message in source_entrypoint.run(parsed_args):
                messages.append(message)
        except Exception as exception:
            if not expecting_exception:
                print("Printing unexpected error from entrypoint_wrapper")
                print("".join(traceback.format_exception(None, exception, exception.__traceback__)))
            uncaught_exception = exception

        captured_logs = log_capture_buffer.getvalue().split("\n")[:-1]

        parent_logger.removeHandler(stream_handler)

        return EntrypointOutput(messages + captured_logs, uncaught_exception)


def make_file(path: Path, file_contents: Optional[Union[str, Mapping[str, Any], List[Mapping[str, Any]]]]) -> str:
    if isinstance(file_contents, str):
        path.write_text(file_contents)
    else:
        path.write_text(json.dumps(file_contents))
    return str(path)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from unittest.mock import MagicMock

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, AirbyteTraceMessage, Level, SyncMode, TraceType, Type
from airbyte_cdk.sources.declarative.declarative_stream import DeclarativeStream

SLICE_NOT_CONSIDERED_FOR_EQUALITY = {}


def test_declarative_stream():
    name = "stream"
    primary_key = "pk"
    cursor_field = "created_at"

    schema_loader = MagicMock()
    json_schema = {"name": {"type": "string"}}
    schema_loader.get_json_schema.return_value = json_schema

    state = MagicMock()
    records = [
        {"pk": 1234, "field": "value"},
        {"pk": 4567, "field": "different_value"},
        AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message="This is a log  message")),
        AirbyteMessage(type=Type.TRACE, trace=AirbyteTraceMessage(type=TraceType.ERROR, emitted_at=12345)),
    ]
    stream_slices = [
        {"date": "2021-01-01"},
        {"date": "2021-01-02"},
        {"date": "2021-01-03"},
    ]

    retriever = MagicMock()
    retriever.state = state
    retriever.read_records.return_value = records
    retriever.stream_slices.return_value = stream_slices

    config = {"api_key": "open_sesame"}

    stream = DeclarativeStream(
        name=name,
        primary_key=primary_key,
        stream_cursor_field="{{ parameters['cursor_field'] }}",
        schema_loader=schema_loader,
        retriever=retriever,
        config=config,
        parameters={"cursor_field": "created_at"},
    )

    assert stream.name == name
    assert stream.get_json_schema() == json_schema
    assert stream.state == state
    input_slice = stream_slices[0]
    assert list(stream.read_records(SyncMode.full_refresh, cursor_field, input_slice, state)) == records
    assert stream.primary_key == primary_key
    assert stream.cursor_field == cursor_field
    assert stream.stream_slices(sync_mode=SyncMode.incremental, cursor_field=cursor_field, stream_state=None) == stream_slices


def test_state_checkpoint_interval():
    stream = DeclarativeStream(
        name="any name",
        primary_key="any primary key",
        stream_cursor_field="{{ parameters['cursor_field'] }}",
        schema_loader=MagicMock(),
        retriever=MagicMock(),
        config={},
        parameters={},
    )

    assert stream.state_checkpoint_interval is None



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import List, Union

from airbyte_cdk.sources.config import BaseConfig
from pydantic import BaseModel, Field


class InnerClass(BaseModel):
    field1: str
    field2: int


class Choice1(BaseModel):
    selected_strategy = Field("option1", const=True)

    name: str
    count: int


class Choice2(BaseModel):
    selected_strategy = Field("option2", const=True)

    sequence: List[str]


class SomeSourceConfig(BaseConfig):
    class Config:
        title = "Some Source"

    items: List[InnerClass]
    choice: Union[Choice1, Choice2]


class TestBaseConfig:
    EXPECTED_SCHEMA = {
        "properties": {
            "choice": {
                "oneOf": [
                    {
                        "properties": {
                            "count": {"title": "Count", "type": "integer"},
                            "name": {"title": "Name", "type": "string"},
                            "selected_strategy": {
                                "const": "option1",
                                "title": "Selected " "Strategy",
                                "type": "string",
                                "default": "option1",
                            },
                        },
                        "required": ["name", "count"],
                        "title": "Choice1",
                        "type": "object",
                    },
                    {
                        "properties": {
                            "selected_strategy": {
                                "const": "option2",
                                "title": "Selected " "Strategy",
                                "type": "string",
                                "default": "option2",
                            },
                            "sequence": {"items": {"type": "string"}, "title": "Sequence", "type": "array"},
                        },
                        "required": ["sequence"],
                        "title": "Choice2",
                        "type": "object",
                    },
                ],
                "title": "Choice",
            },
            "items": {
                "items": {
                    "properties": {"field1": {"title": "Field1", "type": "string"}, "field2": {"title": "Field2", "type": "integer"}},
                    "required": ["field1", "field2"],
                    "title": "InnerClass",
                    "type": "object",
                },
                "title": "Items",
                "type": "array",
            },
        },
        "required": ["items", "choice"],
        "title": "Some Source",
        "type": "object",
    }

    def test_schema_postprocessing(self):
        schema = SomeSourceConfig.schema()
        assert schema == self.EXPECTED_SCHEMA



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from unittest.mock import patch

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.incremental.per_partition_cursor import PerPartitionStreamSlice
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever
from airbyte_cdk.sources.declarative.types import Record

CURSOR_FIELD = "cursor_field"
SYNC_MODE = SyncMode.incremental


class ManifestBuilder:
    def __init__(self):
        self._incremental_sync = None
        self._partition_router = None

    def with_list_partition_router(self, cursor_field, partitions):
        self._partition_router = {
            "type": "ListPartitionRouter",
            "cursor_field": cursor_field,
            "values": partitions,
        }
        return self

    def with_incremental_sync(self, start_datetime, end_datetime, datetime_format, cursor_field, step, cursor_granularity):
        self._incremental_sync = {
            "type": "DatetimeBasedCursor",
            "start_datetime": start_datetime,
            "end_datetime": end_datetime,
            "datetime_format": datetime_format,
            "cursor_field": cursor_field,
            "step": step,
            "cursor_granularity": cursor_granularity,
        }
        return self

    def build(self):
        manifest = {
            "version": "0.34.2",
            "type": "DeclarativeSource",
            "check": {"type": "CheckStream", "stream_names": ["Rates"]},
            "streams": [
                {
                    "type": "DeclarativeStream",
                    "name": "Rates",
                    "primary_key": [],
                    "schema_loader": {
                        "type": "InlineSchemaLoader",
                        "schema": {"$schema": "http://json-schema.org/schema#", "properties": {}, "type": "object"},
                    },
                    "retriever": {
                        "type": "SimpleRetriever",
                        "requester": {
                            "type": "HttpRequester",
                            "url_base": "https://api.apilayer.com",
                            "path": "/exchangerates_data/latest",
                            "http_method": "GET",
                        },
                        "record_selector": {"type": "RecordSelector", "extractor": {"type": "DpathExtractor", "field_path": []}},
                    },
                }
            ],
            "spec": {
                "connection_specification": {
                    "$schema": "http://json-schema.org/draft-07/schema#",
                    "type": "object",
                    "required": [],
                    "properties": {},
                    "additionalProperties": True,
                },
                "documentation_url": "https://example.org",
                "type": "Spec",
            },
        }
        if self._incremental_sync:
            manifest["streams"][0]["incremental_sync"] = self._incremental_sync
        if self._partition_router:
            manifest["streams"][0]["retriever"]["partition_router"] = self._partition_router
        return manifest


def test_given_state_for_only_some_partition_when_stream_slices_then_create_slices_using_state_or_start_from_start_datetime():
    source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_list_partition_router("partition_field", ["1", "2"])
        .with_incremental_sync(
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )
    stream_instance = source.streams({})[0]
    stream_instance.state = {
        "states": [
            {
                "partition": {"partition_field": "1"},
                "cursor": {CURSOR_FIELD: "2022-02-01"},
            }
        ]
    }

    slices = stream_instance.stream_slices(
        sync_mode=SYNC_MODE,
        stream_state={},
    )

    assert list(slices) == [
        {"partition_field": "1", "start_time": "2022-02-01", "end_time": "2022-02-28"},
        {"partition_field": "2", "start_time": "2022-01-01", "end_time": "2022-01-31"},
        {"partition_field": "2", "start_time": "2022-02-01", "end_time": "2022-02-28"},
    ]


def test_given_record_for_partition_when_read_then_update_state():
    source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_list_partition_router("partition_field", ["1", "2"])
        .with_incremental_sync(
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )
    stream_instance = source.streams({})[0]
    list(stream_instance.stream_slices(sync_mode=SYNC_MODE))

    stream_slice = PerPartitionStreamSlice({"partition_field": "1"}, {"start_time": "2022-01-01", "end_time": "2022-01-31"})
    with patch.object(
        SimpleRetriever, "_read_pages", side_effect=[[Record({"a record key": "a record value", CURSOR_FIELD: "2022-01-15"}, stream_slice)]]
    ):
        list(
            stream_instance.read_records(
                sync_mode=SYNC_MODE,
                stream_slice=stream_slice,
                stream_state={"states": []},
                cursor_field=CURSOR_FIELD,
            )
        )

    assert stream_instance.state == {
        "states": [
            {
                "partition": {"partition_field": "1"},
                "cursor": {CURSOR_FIELD: "2022-01-31"},
            }
        ]
    }



from .abstract_file_based_availability_strategy import (
    AbstractFileBasedAvailabilityStrategy,
    AbstractFileBasedAvailabilityStrategyWrapper,
)
from .default_file_based_availability_strategy import DefaultFileBasedAvailabilityStrategy

__all__ = ["AbstractFileBasedAvailabilityStrategy", "AbstractFileBasedAvailabilityStrategyWrapper", "DefaultFileBasedAvailabilityStrategy"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import datetime

import pytest
from airbyte_cdk.sources.declarative.interpolation.jinja import JinjaInterpolation
from freezegun import freeze_time
from jinja2.exceptions import TemplateSyntaxError

interpolation = JinjaInterpolation()


def test_get_value_from_config():
    s = "{{ config['date'] }}"
    config = {"date": "2022-01-01"}
    val = interpolation.eval(s, config)
    assert val == "2022-01-01"


@pytest.mark.parametrize(
    "valid_types, expected_value",
    [
        pytest.param((str,), "1234J", id="test_value_is_a_string_if_valid_types_is_str"),
        pytest.param(None, 1234j, id="test_value_is_interpreted_as_complex_number_by_default"),
    ],
)
def test_get_value_with_complex_number(valid_types, expected_value):
    s = "{{ config['value'] }}"
    config = {"value": "1234J"}
    val = interpolation.eval(s, config, valid_types=valid_types)
    assert val == expected_value


def test_get_value_from_stream_slice():
    s = "{{ stream_slice['date'] }}"
    config = {"date": "2022-01-01"}
    stream_slice = {"date": "2020-09-09"}
    val = interpolation.eval(s, config, **{"stream_slice": stream_slice})
    assert val == "2020-09-09"


def test_get_value_from_a_list_of_mappings():
    s = "{{ records[0]['date'] }}"
    config = {"date": "2022-01-01"}
    records = [{"date": "2020-09-09"}]
    val = interpolation.eval(s, config, **{"records": records})
    assert val == "2020-09-09"


@pytest.mark.parametrize(
    "s, value",
    [
        pytest.param("{{1}}", 1, id="test_number"),
        pytest.param("{{1}}", 1, id="test_number"),
        pytest.param("{{[1,2]}}", [1, 2], id="test_list"),
        pytest.param("{{ {1:2} }}", {1: 2}, id="test_dict"),
        pytest.param("{{ 1+2 }}", 3, id="test_addition"),
    ],
)
def test_literals(s, value):
    val = interpolation.eval(s, None)
    assert val == value


@pytest.mark.parametrize(
    "context, input_string, expected_value",
    [
        pytest.param(
            {"stream_slice": {"stream_slice_key": "hello"}},
            "{{ stream_slice['stream_slice_key'] }}",
            "hello",
            id="test_get_value_from_stream_slice",
        ),
        pytest.param(
            {"stream_slice": {"stream_slice_key": "hello"}},
            "{{ stream_partition['stream_slice_key'] }}",
            "hello",
            id="test_get_value_from_stream_slicer",
        ),
        pytest.param(
            {"stream_slice": {"stream_slice_key": "hello"}},
            "{{ stream_interval['stream_slice_key'] }}",
            "hello",
            id="test_get_value_from_stream_interval",
        ),
    ],
)
def test_stream_slice_alias(context, input_string, expected_value):
    config = {}
    val = interpolation.eval(input_string, config, **context)
    assert val == expected_value


@pytest.mark.parametrize(
    "alias",
    [
        pytest.param("stream_interval", id="test_error_is_raised_if_stream_interval_in_context"),
        pytest.param("stream_partition", id="test_error_is_raised_if_stream_partition_in_context"),
    ],
)
def test_error_is_raised_if_alias_is_already_in_context(alias):
    config = {}
    context = {alias: "a_value"}
    with pytest.raises(ValueError):
        interpolation.eval("a_key", config, **context)


def test_positive_day_delta():
    delta_template = "{{ day_delta(25) }}"
    interpolation = JinjaInterpolation()
    val = interpolation.eval(delta_template, {})

    # We need to assert against an earlier delta since the interpolation function runs datetime.now() a few milliseconds earlier
    assert val > (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=24, hours=23)).strftime("%Y-%m-%dT%H:%M:%S.%f%z")


def test_positive_day_delta_with_format():
    delta_template = "{{ day_delta(25,format='%Y-%m-%d') }}"

    with freeze_time("2021-01-01 03:04:05"):
        val = interpolation.eval(delta_template, {})
        assert val == "2021-01-26"


def test_negative_day_delta():
    delta_template = "{{ day_delta(-25) }}"
    val = interpolation.eval(delta_template, {})

    assert val <= (datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=25)).strftime("%Y-%m-%dT%H:%M:%S.%f%z")


@pytest.mark.parametrize(
    "s, expected_value",
    [
        pytest.param("{{ timestamp(1621439283) }}", 1621439283, id="test_timestamp_from_timestamp"),
        pytest.param("{{ timestamp('2021-05-19') }}", 1621382400, id="test_timestamp_from_string"),
        pytest.param("{{ timestamp('2017-01-01T00:00:00.0Z') }}", 1483228800, id="test_timestamp_from_rfc3339"),
        pytest.param("{{ max(1,2) }}", 2, id="test_max"),
    ],
)
def test_macros(s, expected_value):
    config = {}
    val = interpolation.eval(s, config)
    assert val == expected_value


@pytest.mark.parametrize(
    "template_string",
    [
        pytest.param("{{ import os) }}", id="test_jinja_with_import"),
        pytest.param("{{ [a for a in range(1000000000)] }}", id="test_jinja_with_list_comprehension"),
    ],
)
def test_invalid_jinja_statements(template_string):
    config = {"key": "value"}
    with pytest.raises(TemplateSyntaxError):
        interpolation.eval(template_string, config=config)


@pytest.mark.parametrize(
    "template_string",
    [
        # This test stalls if range is removed from JinjaInterpolation.RESTRICTED_BUILTIN_FUNCTIONS
        pytest.param(
            """
       {% set a = 1 %}
       {% set b = 1 %}
       {% for i in range(1000000000) %}
       {% endfor %}
        {{ a }}""",
            id="test_jinja_with_very_long_running_compute",
        ),
        pytest.param("{{ eval ('2+2') }}", id="test_jinja_with_eval"),
        pytest.param("{{ getattr(config, 'key') }}", id="test_getattr"),
        pytest.param("{{ setattr(config, 'password', 'hunter2') }}", id="test_setattr"),
        pytest.param("{{ globals()  }}", id="test_jinja_with_globals"),
        pytest.param("{{ locals()  }}", id="test_jinja_with_globals"),
        pytest.param("{{ eval ('2+2') }}", id="test_jinja_with_eval"),
        pytest.param("{{ eval }}", id="test_jinja_with_eval"),
    ],
)
def test_restricted_builtin_functions_are_not_executed(template_string):
    config = {"key": JinjaInterpolation}
    with pytest.raises(ValueError):
        interpolation.eval(template_string, config=config)


@pytest.mark.parametrize(
    "template_string, expected_value, expected_error",
    [
        pytest.param("{{ to_be }}", "that_is_the_question", None, id="valid_template_variable"),
        pytest.param("{{ missingno }}", None, ValueError, id="undeclared_template_variable"),
        pytest.param("{{ to_be and or_not_to_be }}", None, ValueError, id="one_undeclared_template_variable"),
    ],
)
def test_undeclared_variables(template_string, expected_error, expected_value):
    config = {"key": JinjaInterpolation}

    if expected_error:
        with pytest.raises(expected_error):
            interpolation.eval(template_string, config=config, **{"to_be": "that_is_the_question"})
    else:
        actual_value = interpolation.eval(template_string, config=config, **{"to_be": "that_is_the_question"})
        assert actual_value == expected_value


@freeze_time("2021-09-01")
@pytest.mark.parametrize(
    "template_string, expected_value",
    [
        pytest.param("{{ now_utc() }}", "2021-09-01 00:00:00+00:00", id="test_now_utc"),
        pytest.param("{{ now_utc().strftime('%Y-%m-%d') }}", "2021-09-01", id="test_now_utc_strftime"),
        pytest.param("{{ today_utc() }}", "2021-09-01", id="test_today_utc"),
        pytest.param("{{ today_utc().strftime('%Y/%m/%d') }}", "2021/09/01", id="test_todat_utc_stftime"),
        pytest.param("{{ timestamp(1646006400) }}", 1646006400, id="test_timestamp_from_timestamp"),
        pytest.param("{{ timestamp('2022-02-28') }}", 1646006400, id="test_timestamp_from_timestamp"),
        pytest.param("{{ timestamp('2022-02-28T00:00:00Z') }}", 1646006400, id="test_timestamp_from_timestamp"),
        pytest.param("{{ timestamp('2022-02-28 00:00:00Z') }}", 1646006400, id="test_timestamp_from_timestamp"),
        pytest.param("{{ timestamp('2022-02-28T00:00:00-08:00') }}", 1646035200, id="test_timestamp_from_date_with_tz"),
        pytest.param("{{ max(2, 3) }}", 3, id="test_max_with_arguments"),
        pytest.param("{{ max([2, 3]) }}", 3, id="test_max_with_list"),
        pytest.param("{{ day_delta(1) }}", "2021-09-02T00:00:00.000000+0000", id="test_day_delta"),
        pytest.param("{{ day_delta(-1) }}", "2021-08-31T00:00:00.000000+0000", id="test_day_delta_negative"),
        pytest.param("{{ day_delta(1, format='%Y-%m-%d') }}", "2021-09-02", id="test_day_delta_with_format"),
        pytest.param("{{ duration('P1D') }}", "1 day, 0:00:00", id="test_duration_one_day"),
        pytest.param("{{ duration('P6DT23H') }}", "6 days, 23:00:00", id="test_duration_six_days_and_23_hours"),
        pytest.param(
            "{{ (now_utc() - duration('P1D')).strftime('%Y-%m-%dT%H:%M:%SZ') }}",
            "2021-08-31T00:00:00Z",
            id="test_now_utc_with_duration_and_format",
        ),
    ],
)
def test_macros_examples(template_string, expected_value):
    # The outputs of this test are referenced in declarative_component_schema.yaml
    # If you change the expected output, you must also change the expected output in declarative_component_schema.yaml
    now_utc = interpolation.eval(template_string, {})
    assert now_utc == expected_value



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import logging
from abc import ABC, abstractmethod
from collections import deque
from typing import Callable, Deque, Iterable, List, Optional

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, Level, Type
from airbyte_cdk.sources.utils.types import JsonType
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets

_LOGGER = logging.getLogger("MessageRepository")
_SUPPORTED_MESSAGE_TYPES = {Type.CONTROL, Type.LOG}
LogMessage = dict[str, JsonType]

_SEVERITY_BY_LOG_LEVEL = {
    Level.FATAL: 1,
    Level.ERROR: 2,
    Level.WARN: 3,
    Level.INFO: 4,
    Level.DEBUG: 5,
    Level.TRACE: 5,
}


def _is_severe_enough(threshold: Level, level: Level) -> bool:
    if threshold not in _SEVERITY_BY_LOG_LEVEL:
        _LOGGER.warning(f"Log level {threshold} for threshold is not supported. This is probably a CDK bug. Please contact Airbyte.")
        return True

    if level not in _SEVERITY_BY_LOG_LEVEL:
        _LOGGER.warning(
            f"Log level {level} is not supported. This is probably a source bug. Please contact the owner of the source or Airbyte."
        )
        return True

    return _SEVERITY_BY_LOG_LEVEL[threshold] >= _SEVERITY_BY_LOG_LEVEL[level]


class MessageRepository(ABC):
    @abstractmethod
    def emit_message(self, message: AirbyteMessage) -> None:
        raise NotImplementedError()

    @abstractmethod
    def log_message(self, level: Level, message_provider: Callable[[], LogMessage]) -> None:
        """
        Computing messages can be resource consuming. This method is specialized for logging because we want to allow for lazy evaluation if
        the log level is less severe than what is configured
        """
        raise NotImplementedError()

    @abstractmethod
    def consume_queue(self) -> Iterable[AirbyteMessage]:
        raise NotImplementedError()


class NoopMessageRepository(MessageRepository):
    def emit_message(self, message: AirbyteMessage) -> None:
        pass

    def log_message(self, level: Level, message_provider: Callable[[], LogMessage]) -> None:
        pass

    def consume_queue(self) -> Iterable[AirbyteMessage]:
        return []


class InMemoryMessageRepository(MessageRepository):
    def __init__(self, log_level: Level = Level.INFO) -> None:
        self._message_queue: Deque[AirbyteMessage] = deque()
        self._log_level = log_level

    def emit_message(self, message: AirbyteMessage) -> None:
        self._message_queue.append(message)

    def log_message(self, level: Level, message_provider: Callable[[], LogMessage]) -> None:
        if _is_severe_enough(self._log_level, level):
            self.emit_message(
                AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=level, message=filter_secrets(json.dumps(message_provider()))))
            )

    def consume_queue(self) -> Iterable[AirbyteMessage]:
        while self._message_queue:
            yield self._message_queue.popleft()


class LogAppenderMessageRepositoryDecorator(MessageRepository):
    def __init__(self, dict_to_append: LogMessage, decorated: MessageRepository, log_level: Level = Level.INFO):
        self._dict_to_append = dict_to_append
        self._decorated = decorated
        self._log_level = log_level

    def emit_message(self, message: AirbyteMessage) -> None:
        self._decorated.emit_message(message)

    def log_message(self, level: Level, message_provider: Callable[[], LogMessage]) -> None:
        if _is_severe_enough(self._log_level, level):
            message = message_provider()
            self._append_second_to_first(message, self._dict_to_append)
            self._decorated.log_message(level, lambda: message)

    def consume_queue(self) -> Iterable[AirbyteMessage]:
        return self._decorated.consume_queue()

    def _append_second_to_first(self, first: LogMessage, second: LogMessage, path: Optional[List[str]] = None) -> LogMessage:
        if path is None:
            path = []

        for key in second:
            if key in first:
                if isinstance(first[key], dict) and isinstance(second[key], dict):
                    self._append_second_to_first(first[key], second[key], path + [str(key)])  # type: ignore # type is verified above
                else:
                    if first[key] != second[key]:
                        _LOGGER.warning("Conflict at %s" % ".".join(path + [str(key)]))
                    first[key] = second[key]
            else:
                first[key] = second[key]
        return first



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import asyncio
from datetime import datetime
from unittest import mock
from unittest.mock import MagicMock, call, mock_open, patch

import pytest
import requests
from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.config.unstructured_format import APIParameterConfigModel, APIProcessingConfigModel, UnstructuredFormat
from airbyte_cdk.sources.file_based.exceptions import RecordParseError
from airbyte_cdk.sources.file_based.file_types import UnstructuredParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.utils.traced_exception import AirbyteTracedException
from unstructured.documents.elements import ElementMetadata, Formula, ListItem, Text, Title
from unstructured.file_utils.filetype import FileType

FILE_URI = "path/to/file.xyz"


@pytest.mark.parametrize(
    "filetype, format_config, raises",
    [
        pytest.param(
            FileType.MD,
            UnstructuredFormat(skip_unprocessable_files=False),
            False,
            id="markdown_file",
        ),
        pytest.param(
            FileType.CSV,
            UnstructuredFormat(skip_unprocessable_files=False),
            True,
            id="wrong_file_format",
        ),
        pytest.param(
            FileType.CSV,
            UnstructuredFormat(skip_unprocessable_files=True),
            False,
            id="wrong_file_format_skipping",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_files=False),
            False,
            id="pdf_file",
        ),
        pytest.param(
            FileType.DOCX,
            UnstructuredFormat(skip_unprocessable_files=False),
            False,
            id="docx_file",
        ),
        pytest.param(
            FileType.PPTX,
            UnstructuredFormat(skip_unprocessable_files=False),
            False,
            id="pptx_file",
        ),
    ],
)
@patch("airbyte_cdk.sources.file_based.file_types.unstructured_parser.detect_filetype")
def test_infer_schema(mock_detect_filetype, filetype, format_config, raises):
    # use a fresh event loop to avoid leaking into other tests
    main_loop = asyncio.get_event_loop()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    stream_reader = MagicMock()
    mock_open(stream_reader.open_file)
    fake_file = MagicMock()
    fake_file.uri = FILE_URI
    logger = MagicMock()
    mock_detect_filetype.return_value = filetype
    config = MagicMock()
    config.format = format_config
    if raises:
        with pytest.raises(RecordParseError):
            loop.run_until_complete(UnstructuredParser().infer_schema(config, fake_file, stream_reader, logger))
    else:
        schema = loop.run_until_complete(UnstructuredParser().infer_schema(config, MagicMock(), MagicMock(), MagicMock()))
        assert schema == {
            "content": {"type": "string", "description": "Content of the file as markdown. Might be null if the file could not be parsed"},
            "document_key": {"type": "string", "description": "Unique identifier of the document, e.g. the file path"},
            "_ab_source_file_parse_error": {"type": "string", "description": "Error message if the file could not be parsed even though the file is supported"},
        }
    loop.close()
    asyncio.set_event_loop(main_loop)


@pytest.mark.parametrize(
    "filetype, format_config, parse_result, raises, expected_records, parsing_error",
    [
        pytest.param(
            FileType.MD,
            UnstructuredFormat(skip_unprocessable_files=False),
            "test",
            False,
            [
                {
                    "content": "test",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None,
                }
            ],
            False,
            id="markdown_file",
        ),
        pytest.param(
            FileType.CSV,
            UnstructuredFormat(skip_unprocessable_files=False),
            None,
            True,
            None,
            False,
            id="wrong_file_format",
        ),
        pytest.param(
            FileType.CSV,
            UnstructuredFormat(skip_unprocessable_files=True),
            None,
            False,
            [
                {
                    "content": None,
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable. Contact Support if you need assistance.\nfilename=path/to/file.xyz message=File type FileType.CSV is not supported. Supported file types are FileType.MD, FileType.PDF, FileType.DOCX, FileType.PPTX, FileType.TXT",
                }
            ],
            False,
            id="skip_unprocessable_files",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_files=False),
            [
                Title("heading"),
                Text("This is the text"),
                ListItem("This is a list item"),
                Formula("This is a formula"),
            ],
            False,
            [
                {
                    "content": "# heading\n\nThis is the text\n\n- This is a list item\n\n```\nThis is a formula\n```",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None,
                }
            ],
            False,
            id="pdf_file",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_files=False),
            [
                Title("first level heading", metadata=ElementMetadata(category_depth=1)),
                Title("second level heading", metadata=ElementMetadata(category_depth=2)),
            ],
            False,
            [
                {
                    "content": "# first level heading\n\n## second level heading",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None,
                }
            ],
            False,
            id="multi_level_headings",
        ),
        pytest.param(
            FileType.DOCX,
            UnstructuredFormat(skip_unprocessable_files=False),
            [
                Title("heading"),
                Text("This is the text"),
                ListItem("This is a list item"),
                Formula("This is a formula"),
            ],
            False,
            [
                {
                    "content": "# heading\n\nThis is the text\n\n- This is a list item\n\n```\nThis is a formula\n```",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None,
                }
            ],
            False,
            id="docx_file",
        ),
        pytest.param(
            FileType.DOCX,
            UnstructuredFormat(skip_unprocessable_files=True),
            "",
            False,
            [
                {
                    "content": None,
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable. Contact Support if you need assistance.\nfilename=path/to/file.xyz message=weird parsing error"
                }
            ],
            True,
            id="exception_during_parsing",
        ),
    ],
)
@patch("unstructured.partition.pdf.partition_pdf")
@patch("unstructured.partition.pptx.partition_pptx")
@patch("unstructured.partition.docx.partition_docx")
@patch("airbyte_cdk.sources.file_based.file_types.unstructured_parser.detect_filetype")
def test_parse_records(
    mock_detect_filetype,
    mock_partition_docx,
    mock_partition_pptx,
    mock_partition_pdf,
    filetype,
    format_config,
    parse_result,
    raises,
    expected_records,
    parsing_error,
):
    stream_reader = MagicMock()
    mock_open(stream_reader.open_file, read_data=bytes(str(parse_result), "utf-8"))
    fake_file = RemoteFile(uri=FILE_URI, last_modified=datetime.now())
    fake_file.uri = FILE_URI
    logger = MagicMock()
    config = MagicMock()
    config.format = format_config
    mock_detect_filetype.return_value = filetype
    if parsing_error:
        mock_partition_docx.side_effect = Exception("weird parsing error")
        mock_partition_pptx.side_effect = Exception("weird parsing error")
        mock_partition_pdf.side_effect = Exception("weird parsing error")
    else:
        mock_partition_docx.return_value = parse_result
        mock_partition_pptx.return_value = parse_result
        mock_partition_pdf.return_value = parse_result
    if raises:
        with pytest.raises(RecordParseError):
            list(UnstructuredParser().parse_records(config, fake_file, stream_reader, logger, MagicMock()))
    else:
        assert list(UnstructuredParser().parse_records(config, fake_file, stream_reader, logger, MagicMock())) == expected_records


@pytest.mark.parametrize(
    "format_config, raises_for_status, json_response, is_ok, expected_error",
    [
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False),
            False,
            {"status": "ok"},
            True,
            None,
            id="local",
        ),
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False, strategy="fast"),
            False,
            {"status": "ok"},
            True,
            None,
            id="local_ok_strategy",
        ),
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False, strategy="hi_res"),
            False,
            {"status": "ok"},
            False,
            "Hi-res strategy is not supported for local processing",
            id="local_unsupported_strategy",
        ),
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            False,
            [{"type": "Title", "text": "Airbyte source connection test"}],
            True,
            None,
            id="api_ok",
        ),
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            True,
            None,
            False,
            "API error",
            id="api_error",
        ),
        pytest.param(
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            False,
            {"unexpected": "response"},
            False,
            "Error",
            id="unexpected_handling_error",
        ),
    ],
)
@patch("airbyte_cdk.sources.file_based.file_types.unstructured_parser.requests")
def test_check_config(requests_mock, format_config, raises_for_status, json_response, is_ok, expected_error):
    mock_response = MagicMock()
    mock_response.json.return_value = json_response
    if raises_for_status:
        mock_response.raise_for_status.side_effect = Exception("API error")
    requests_mock.post.return_value = mock_response
    result, error = UnstructuredParser().check_config(FileBasedStreamConfig(name="test", format=format_config))
    assert result == is_ok
    if expected_error:
        assert expected_error in error


@pytest.mark.parametrize(
    "filetype, format_config, raises_for_status, file_content, json_response, expected_requests, raises, expected_records, http_status_code",
    [
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            None,
            "test",
            [{"type": "Text", "text": "test"}],
            [call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")})],
            False,
            [
                {
                    "content": "test",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None
                }
            ],
            200,
            id="basic_request",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, strategy="hi_res", processing=APIProcessingConfigModel(mode="api", api_key="test", api_url="http://localhost:8000", parameters=[APIParameterConfigModel(name="include_page_breaks", value="true"), APIParameterConfigModel(name="ocr_languages", value="eng"), APIParameterConfigModel(name="ocr_languages", value="kor")])),
            None,
            "test",
            [{"type": "Text", "text": "test"}],
            [call("http://localhost:8000/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "hi_res", "include_page_breaks": "true", "ocr_languages": ["eng", "kor"]}, files={"files": ("filename", mock.ANY, "application/pdf")})],
            False,
            [
                {
                    "content": "test",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None
                }
            ],
            200,
            id="request_with_params",
        ),
        pytest.param(
            FileType.MD,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            None,
            "# Mymarkdown",
            None,
            None,
            False,
            [
                {
                    "content": "# Mymarkdown",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None
                }
            ],
            200,
            id="handle_markdown_locally",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            [
                requests.exceptions.RequestException("API error"),
                requests.exceptions.RequestException("API error"),
                requests.exceptions.RequestException("API error"),
                requests.exceptions.RequestException("API error"),
                requests.exceptions.RequestException("API error"),
            ],
            "test",
            None,
            [
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
            ],
            True,
            None,
            200,
            id="retry_and_raise_on_api_error",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            [
                requests.exceptions.RequestException("API error"),
                requests.exceptions.RequestException("API error"),
                None,
            ],
            "test",
            [{"type": "Text", "text": "test"}],
            [
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
            ],
            False,
            [
                {
                    "content": "test",
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": None
                }
            ],
            200,
            id="retry_and_recover",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            [
                Exception("Unexpected error"),
            ],
            "test",
            [{"type": "Text", "text": "test"}],
            [
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
            ],
            True,
            None,
            200,
            id="no_retry_on_unexpected_error",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            [
                requests.exceptions.RequestException("API error", response=MagicMock(status_code=400)),
            ],
            "test",
            [{"type": "Text", "text": "test"}],
            [
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
                call().raise_for_status(),
            ],
            True,
            None,
            400,
            id="no_retry_on_400_error",
        ),
        pytest.param(
            FileType.PDF,
            UnstructuredFormat(skip_unprocessable_file_types=False, processing=APIProcessingConfigModel(mode="api", api_key="test")),
            None,
            "test",
            [{"detail": "Something went wrong"}],
            [
                call("https://api.unstructured.io/general/v0/general", headers={"accept": "application/json", "unstructured-api-key": "test"}, data={"strategy": "auto"}, files={"files": ("filename", mock.ANY, "application/pdf")}),
            ],
            False,
            [
                {
                    "content": None,
                    "document_key": FILE_URI,
                    "_ab_source_file_parse_error": "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable. Contact Support if you need assistance.\nfilename=path/to/file.xyz message=[{'detail': 'Something went wrong'}]",
                }
            ],
            422,
            id="error_record_on_422_error",
        ),
    ],
)
@patch("airbyte_cdk.sources.file_based.file_types.unstructured_parser.requests")
@patch("airbyte_cdk.sources.file_based.file_types.unstructured_parser.detect_filetype")
@patch('time.sleep', side_effect=lambda _: None)
def test_parse_records_remotely(
    time_mock,
    mock_detect_filetype,
    requests_mock,
    filetype,
    format_config,
    raises_for_status,
    file_content,
    json_response,
    expected_requests,
    raises,
    expected_records,
    http_status_code
):
    stream_reader = MagicMock()
    mock_open(stream_reader.open_file, read_data=bytes(str(file_content), "utf-8"))
    fake_file = RemoteFile(uri=FILE_URI, last_modified=datetime.now())
    fake_file.uri = FILE_URI
    logger = MagicMock()
    config = MagicMock()
    config.format = format_config
    mock_detect_filetype.return_value = filetype
    mock_response = MagicMock()
    mock_response.json.return_value = json_response
    mock_response.status_code = http_status_code
    if raises_for_status:
        mock_response.raise_for_status.side_effect = raises_for_status
    requests_mock.post.return_value = mock_response
    requests_mock.exceptions.RequestException = requests.exceptions.RequestException

    if raises:
        with pytest.raises(AirbyteTracedException) as exc:
            list(UnstructuredParser().parse_records(config, fake_file, stream_reader, logger, MagicMock()))
        # Failures from the API are treated as config errors
        assert exc.value.failure_type == FailureType.config_error
    else:
        assert list(UnstructuredParser().parse_records(config, fake_file, stream_reader, logger, MagicMock())) == expected_records

    if expected_requests:
        requests_mock.post.assert_has_calls(expected_requests)
    else:
        requests_mock.post.assert_not_called()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import copy
import dataclasses
import json
import logging
import os
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
import requests
from airbyte_cdk import connector_builder
from airbyte_cdk.connector_builder.connector_builder_handler import (
    DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE,
    DEFAULT_MAXIMUM_NUMBER_OF_SLICES,
    DEFAULT_MAXIMUM_RECORDS,
    TestReadLimits,
    create_source,
    get_limits,
    resolve_manifest,
)
from airbyte_cdk.connector_builder.main import handle_connector_builder_request, handle_request, read_stream
from airbyte_cdk.connector_builder.models import LogMessage, StreamRead, StreamReadPages, StreamReadSlices
from airbyte_cdk.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteStream,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    ConnectorSpecification,
    DestinationSyncMode,
    Level,
    SyncMode,
)
from airbyte_cdk.models import Type
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.declarative.declarative_stream import DeclarativeStream
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.retrievers import SimpleRetrieverTestReadDecorator
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever
from unit_tests.connector_builder.utils import create_configured_catalog

_stream_name = "stream_with_custom_requester"
_stream_primary_key = "id"
_stream_url_base = "https://api.sendgrid.com"
_stream_options = {"name": _stream_name, "primary_key": _stream_primary_key, "url_base": _stream_url_base}
_page_size = 2

MANIFEST = {
    "version": "0.30.3",
    "definitions": {
        "retriever": {
            "paginator": {
                "type": "DefaultPaginator",
                "page_size": _page_size,
                "page_size_option": {"inject_into": "request_parameter", "field_name": "page_size"},
                "page_token_option": {"inject_into": "path", "type": "RequestPath"},
                "pagination_strategy": {
                    "type": "CursorPagination",
                    "cursor_value": "{{ response._metadata.next }}",
                    "page_size": _page_size,
                },
            },
            "partition_router": {
                "type": "ListPartitionRouter",
                "values": ["0", "1", "2", "3", "4", "5", "6", "7"],
                "cursor_field": "item_id",
            },
            ""
            "requester": {
                "path": "/v3/marketing/lists",
                "authenticator": {"type": "BearerAuthenticator", "api_token": "{{ config.apikey }}"},
                "request_parameters": {"a_param": "10"},
            },
            "record_selector": {"extractor": {"field_path": ["result"]}},
        },
    },
    "streams": [
        {
            "type": "DeclarativeStream",
            "$parameters": _stream_options,
            "retriever": "#/definitions/retriever",
        },
    ],
    "check": {"type": "CheckStream", "stream_names": ["lists"]},
    "spec": {
        "connection_specification": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": True,
        },
        "type": "Spec",
    },
}

OAUTH_MANIFEST = {
    "version": "0.30.3",
    "definitions": {
        "retriever": {
            "paginator": {
                "type": "DefaultPaginator",
                "page_size": _page_size,
                "page_size_option": {"inject_into": "request_parameter", "field_name": "page_size"},
                "page_token_option": {"inject_into": "path", "type": "RequestPath"},
                "pagination_strategy": {
                    "type": "CursorPagination",
                    "cursor_value": "{{ response._metadata.next }}",
                    "page_size": _page_size,
                },
            },
            "partition_router": {
                "type": "ListPartitionRouter",
                "values": ["0", "1", "2", "3", "4", "5", "6", "7"],
                "cursor_field": "item_id",
            },
            ""
            "requester": {
                "path": "/v3/marketing/lists",
                "authenticator": {"type": "OAuthAuthenticator", "api_token": "{{ config.apikey }}"},
                "request_parameters": {"a_param": "10"},
            },
            "record_selector": {"extractor": {"field_path": ["result"]}},
        },
    },
    "streams": [
        {
            "type": "DeclarativeStream",
            "$parameters": _stream_options,
            "retriever": "#/definitions/retriever",
        },
    ],
    "check": {"type": "CheckStream", "stream_names": ["lists"]},
    "spec": {
        "connection_specification": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": True,
        },
        "type": "Spec",
    },
}

RESOLVE_MANIFEST_CONFIG = {
    "__injected_declarative_manifest": MANIFEST,
    "__command": "resolve_manifest",
}

TEST_READ_CONFIG = {
    "__injected_declarative_manifest": MANIFEST,
    "__command": "test_read",
    "__test_read_config": {"max_pages_per_slice": 2, "max_slices": 5, "max_records": 10},
}

DUMMY_CATALOG = {
    "streams": [
        {
            "stream": {
                "name": "dummy_stream",
                "json_schema": {"$schema": "http://json-schema.org/draft-07/schema#", "type": "object", "properties": {}},
                "supported_sync_modes": ["full_refresh"],
                "source_defined_cursor": False,
            },
            "sync_mode": "full_refresh",
            "destination_sync_mode": "overwrite",
        }
    ]
}

CONFIGURED_CATALOG = {
    "streams": [
        {
            "stream": {
                "name": _stream_name,
                "json_schema": {"$schema": "http://json-schema.org/draft-07/schema#", "type": "object", "properties": {}},
                "supported_sync_modes": ["full_refresh"],
                "source_defined_cursor": False,
            },
            "sync_mode": "full_refresh",
            "destination_sync_mode": "overwrite",
        }
    ]
}

MOCK_RESPONSE = {
    "result": [
        {"id": 1, "name": "Nora Moon", "position": "director"},
        {"id": 2, "name": "Hae Sung Jung", "position": "cinematographer"},
        {"id": 3, "name": "Arthur Zenneranski", "position": "composer"},
    ]
}


@pytest.fixture
def valid_resolve_manifest_config_file(tmp_path):
    config_file = tmp_path / "config.json"
    config_file.write_text(json.dumps(RESOLVE_MANIFEST_CONFIG))
    return config_file


@pytest.fixture
def valid_read_config_file(tmp_path):
    config_file = tmp_path / "config.json"
    config_file.write_text(json.dumps(TEST_READ_CONFIG))
    return config_file


@pytest.fixture
def dummy_catalog(tmp_path):
    config_file = tmp_path / "catalog.json"
    config_file.write_text(json.dumps(DUMMY_CATALOG))
    return config_file


@pytest.fixture
def configured_catalog(tmp_path):
    config_file = tmp_path / "catalog.json"
    config_file.write_text(json.dumps(CONFIGURED_CATALOG))
    return config_file


@pytest.fixture
def invalid_config_file(tmp_path):
    invalid_config = copy.deepcopy(RESOLVE_MANIFEST_CONFIG)
    invalid_config["__command"] = "bad_command"
    config_file = tmp_path / "config.json"
    config_file.write_text(json.dumps(invalid_config))
    return config_file


def _mocked_send(self, request, **kwargs) -> requests.Response:
    """
    Mocks the outbound send operation to provide faster and more reliable responses compared to actual API requests
    """
    response = requests.Response()
    response.request = request
    response.status_code = 200
    response.headers = {"header": "value"}
    response_body = MOCK_RESPONSE
    response._content = json.dumps(response_body).encode("utf-8")
    return response


def test_handle_resolve_manifest(valid_resolve_manifest_config_file, dummy_catalog):
    with mock.patch.object(connector_builder.main, "handle_connector_builder_request") as patched_handle:
        handle_request(["read", "--config", str(valid_resolve_manifest_config_file), "--catalog", str(dummy_catalog)])
        assert patched_handle.call_count == 1


def test_handle_test_read(valid_read_config_file, configured_catalog):
    with mock.patch.object(connector_builder.main, "handle_connector_builder_request") as patch:
        handle_request(["read", "--config", str(valid_read_config_file), "--catalog", str(configured_catalog)])
        assert patch.call_count == 1


def test_resolve_manifest(valid_resolve_manifest_config_file):
    config = copy.deepcopy(RESOLVE_MANIFEST_CONFIG)
    command = "resolve_manifest"
    config["__command"] = command
    source = ManifestDeclarativeSource(MANIFEST)
    limits = TestReadLimits()
    resolved_manifest = handle_connector_builder_request(source, command, config, create_configured_catalog("dummy_stream"), limits)

    expected_resolved_manifest = {
        "type": "DeclarativeSource",
        "version": "0.30.3",
        "definitions": {
            "retriever": {
                "paginator": {
                    "type": "DefaultPaginator",
                    "page_size": _page_size,
                    "page_size_option": {"inject_into": "request_parameter", "field_name": "page_size"},
                    "page_token_option": {"inject_into": "path", "type": "RequestPath"},
                    "pagination_strategy": {
                        "type": "CursorPagination",
                        "cursor_value": "{{ response._metadata.next }}",
                        "page_size": _page_size,
                    },
                },
                "partition_router": {
                    "type": "ListPartitionRouter",
                    "values": ["0", "1", "2", "3", "4", "5", "6", "7"],
                    "cursor_field": "item_id",
                },
                "requester": {
                    "path": "/v3/marketing/lists",
                    "authenticator": {"type": "BearerAuthenticator", "api_token": "{{ config.apikey }}"},
                    "request_parameters": {"a_param": "10"},
                },
                "record_selector": {"extractor": {"field_path": ["result"]}},
            },
        },
        "streams": [
            {
                "type": "DeclarativeStream",
                "retriever": {
                    "type": "SimpleRetriever",
                    "paginator": {
                        "type": "DefaultPaginator",
                        "page_size": _page_size,
                        "page_size_option": {
                            "type": "RequestOption",
                            "inject_into": "request_parameter",
                            "field_name": "page_size",
                            "name": _stream_name,
                            "primary_key": _stream_primary_key,
                            "url_base": _stream_url_base,
                            "$parameters": _stream_options,
                        },
                        "page_token_option": {
                            "type": "RequestPath",
                            "inject_into": "path",
                            "name": _stream_name,
                            "primary_key": _stream_primary_key,
                            "url_base": _stream_url_base,
                            "$parameters": _stream_options,
                        },
                        "pagination_strategy": {
                            "type": "CursorPagination",
                            "cursor_value": "{{ response._metadata.next }}",
                            "name": _stream_name,
                            "primary_key": _stream_primary_key,
                            "url_base": _stream_url_base,
                            "$parameters": _stream_options,
                            "page_size": _page_size,
                        },
                        "name": _stream_name,
                        "primary_key": _stream_primary_key,
                        "url_base": _stream_url_base,
                        "$parameters": _stream_options,
                    },
                    "requester": {
                        "type": "HttpRequester",
                        "path": "/v3/marketing/lists",
                        "authenticator": {
                            "type": "BearerAuthenticator",
                            "api_token": "{{ config.apikey }}",
                            "name": _stream_name,
                            "primary_key": _stream_primary_key,
                            "url_base": _stream_url_base,
                            "$parameters": _stream_options,
                        },
                        "request_parameters": {"a_param": "10"},
                        "name": _stream_name,
                        "primary_key": _stream_primary_key,
                        "url_base": _stream_url_base,
                        "$parameters": _stream_options,
                    },
                    "partition_router": {
                        "type": "ListPartitionRouter",
                        "values": ["0", "1", "2", "3", "4", "5", "6", "7"],
                        "cursor_field": "item_id",
                        "name": _stream_name,
                        "primary_key": _stream_primary_key,
                        "url_base": _stream_url_base,
                        "$parameters": _stream_options,
                    },
                    "record_selector": {
                        "type": "RecordSelector",
                        "extractor": {
                            "type": "DpathExtractor",
                            "field_path": ["result"],
                            "name": _stream_name,
                            "primary_key": _stream_primary_key,
                            "url_base": _stream_url_base,
                            "$parameters": _stream_options,
                        },
                        "name": _stream_name,
                        "primary_key": _stream_primary_key,
                        "url_base": _stream_url_base,
                        "$parameters": _stream_options,
                    },
                    "name": _stream_name,
                    "primary_key": _stream_primary_key,
                    "url_base": _stream_url_base,
                    "$parameters": _stream_options,
                },
                "name": _stream_name,
                "primary_key": _stream_primary_key,
                "url_base": _stream_url_base,
                "$parameters": _stream_options,
            },
        ],
        "check": {"type": "CheckStream", "stream_names": ["lists"]},
        "spec": {
            "connection_specification": {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "required": [],
                "properties": {},
                "additionalProperties": True,
            },
            "type": "Spec",
        },
    }
    assert resolved_manifest.record.data["manifest"] == expected_resolved_manifest
    assert resolved_manifest.record.stream == "resolve_manifest"


def test_resolve_manifest_error_returns_error_response():
    class MockManifestDeclarativeSource:
        @property
        def resolved_manifest(self):
            raise ValueError

    source = MockManifestDeclarativeSource()
    response = resolve_manifest(source)
    assert "Error resolving manifest" in response.trace.error.message


def test_read():
    config = TEST_READ_CONFIG
    source = ManifestDeclarativeSource(MANIFEST)

    real_record = AirbyteRecordMessage(data={"id": "1234", "key": "value"}, emitted_at=1, stream=_stream_name)
    stream_read = StreamRead(
        logs=[{"message": "here be a log message"}],
        slices=[
            StreamReadSlices(
                pages=[StreamReadPages(records=[real_record], request=None, response=None)],
                slice_descriptor=None,
                state=None,
            )
        ],
        auxiliary_requests=[],
        test_read_limit_reached=False,
        inferred_schema=None,
        inferred_datetime_formats=None,
        latest_config_update={},
    )

    expected_airbyte_message = AirbyteMessage(
        type=MessageType.RECORD,
        record=AirbyteRecordMessage(
            stream=_stream_name,
            data={
                "logs": [{"message": "here be a log message"}],
                "slices": [
                    {"pages": [{"records": [real_record], "request": None, "response": None}], "slice_descriptor": None, "state": None}
                ],
                "test_read_limit_reached": False,
                "auxiliary_requests": [],
                "inferred_schema": None,
                "inferred_datetime_formats": None,
                "latest_config_update": {},
            },
            emitted_at=1,
        ),
    )
    limits = TestReadLimits()
    with patch("airbyte_cdk.connector_builder.message_grouper.MessageGrouper.get_message_groups", return_value=stream_read):
        output_record = handle_connector_builder_request(
            source, "test_read", config, ConfiguredAirbyteCatalog.parse_obj(CONFIGURED_CATALOG), limits
        )
        output_record.record.emitted_at = 1
        assert output_record == expected_airbyte_message


def test_config_update():
    manifest = copy.deepcopy(MANIFEST)
    manifest["definitions"]["retriever"]["requester"]["authenticator"] = {
        "type": "OAuthAuthenticator",
        "token_refresh_endpoint": "https://oauth.endpoint.com/tokens/bearer",
        "client_id": "{{ config['credentials']['client_id'] }}",
        "client_secret": "{{ config['credentials']['client_secret'] }}",
        "refresh_token": "{{ config['credentials']['refresh_token'] }}",
        "refresh_token_updater": {},
    }
    config = copy.deepcopy(TEST_READ_CONFIG)
    config["__injected_declarative_manifest"] = manifest
    config["credentials"] = {
        "client_id": "a client id",
        "client_secret": "a client secret",
        "refresh_token": "a refresh token",
    }
    source = ManifestDeclarativeSource(manifest)

    refresh_request_response = {
        "access_token": "an updated access token",
        "refresh_token": "an updated refresh token",
        "expires_in": 3600,
    }
    with patch(
        "airbyte_cdk.sources.streams.http.requests_native_auth.SingleUseRefreshTokenOauth2Authenticator._get_refresh_access_token_response",
        return_value=refresh_request_response,
    ):
        output = handle_connector_builder_request(
            source, "test_read", config, ConfiguredAirbyteCatalog.parse_obj(CONFIGURED_CATALOG), TestReadLimits()
        )
        assert output.record.data["latest_config_update"]


@patch("traceback.TracebackException.from_exception")
def test_read_returns_error_response(mock_from_exception):
    class MockManifestDeclarativeSource:
        def read(self, logger, config, catalog, state):
            raise ValueError("error_message")

        def spec(self, logger: logging.Logger) -> ConnectorSpecification:
            connector_specification = mock.Mock()
            connector_specification.connectionSpecification = {}
            return connector_specification

        @property
        def check_config_against_spec(self):
            return False

    stack_trace = "a stack trace"
    mock_from_exception.return_value = stack_trace

    source = MockManifestDeclarativeSource()
    limits = TestReadLimits()
    response = read_stream(source, TEST_READ_CONFIG, ConfiguredAirbyteCatalog.parse_obj(CONFIGURED_CATALOG), limits)

    expected_stream_read = StreamRead(
        logs=[LogMessage("error_message - a stack trace", "ERROR")],
        slices=[],
        test_read_limit_reached=False,
        auxiliary_requests=[],
        inferred_schema=None,
        inferred_datetime_formats={},
        latest_config_update=None,
    )

    expected_message = AirbyteMessage(
        type=MessageType.RECORD,
        record=AirbyteRecordMessage(stream=_stream_name, data=dataclasses.asdict(expected_stream_read), emitted_at=1),
    )
    response.record.emitted_at = 1
    assert response == expected_message


@pytest.mark.parametrize(
    "command",
    [
        pytest.param("check", id="test_check_command_error"),
        pytest.param("spec", id="test_spec_command_error"),
        pytest.param("discover", id="test_discover_command_error"),
        pytest.param(None, id="test_command_is_none_error"),
        pytest.param("", id="test_command_is_empty_error"),
    ],
)
def test_invalid_protocol_command(command, valid_resolve_manifest_config_file):
    config = copy.deepcopy(RESOLVE_MANIFEST_CONFIG)
    config["__command"] = "resolve_manifest"
    with pytest.raises(SystemExit):
        handle_request([command, "--config", str(valid_resolve_manifest_config_file), "--catalog", ""])


def test_missing_command(valid_resolve_manifest_config_file):
    with pytest.raises(SystemExit):
        handle_request(["--config", str(valid_resolve_manifest_config_file), "--catalog", ""])


def test_missing_catalog(valid_resolve_manifest_config_file):
    with pytest.raises(SystemExit):
        handle_request(["read", "--config", str(valid_resolve_manifest_config_file)])


def test_missing_config(valid_resolve_manifest_config_file):
    with pytest.raises(SystemExit):
        handle_request(["read", "--catalog", str(valid_resolve_manifest_config_file)])


def test_invalid_config_command(invalid_config_file, dummy_catalog):
    with pytest.raises(ValueError):
        handle_request(["read", "--config", str(invalid_config_file), "--catalog", str(dummy_catalog)])


@pytest.fixture
def manifest_declarative_source():
    return mock.Mock(spec=ManifestDeclarativeSource, autospec=True)


def create_mock_retriever(name, url_base, path):
    http_stream = mock.Mock(spec=SimpleRetriever, autospec=True)
    http_stream.name = name
    http_stream.requester = MagicMock()
    http_stream.requester.get_url_base.return_value = url_base
    http_stream.requester.get_path.return_value = path
    http_stream._paginator_path.return_value = None
    return http_stream


def create_mock_declarative_stream(http_stream):
    declarative_stream = mock.Mock(spec=DeclarativeStream, autospec=True)
    declarative_stream.retriever = http_stream
    return declarative_stream


@pytest.mark.parametrize(
    "test_name, config, expected_max_records, expected_max_slices, expected_max_pages_per_slice",
    [
        (
            "test_no_test_read_config",
            {},
            DEFAULT_MAXIMUM_RECORDS,
            DEFAULT_MAXIMUM_NUMBER_OF_SLICES,
            DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE,
        ),
        (
            "test_no_values_set",
            {"__test_read_config": {}},
            DEFAULT_MAXIMUM_RECORDS,
            DEFAULT_MAXIMUM_NUMBER_OF_SLICES,
            DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE,
        ),
        ("test_values_are_set", {"__test_read_config": {"max_slices": 1, "max_pages_per_slice": 2, "max_records": 3}}, 3, 1, 2),
    ],
)
def test_get_limits(test_name, config, expected_max_records, expected_max_slices, expected_max_pages_per_slice):
    limits = get_limits(config)
    assert limits.max_records == expected_max_records
    assert limits.max_pages_per_slice == expected_max_pages_per_slice
    assert limits.max_slices == expected_max_slices


def test_create_source():
    max_records = 3
    max_pages_per_slice = 2
    max_slices = 1
    limits = TestReadLimits(max_records, max_pages_per_slice, max_slices)

    config = {"__injected_declarative_manifest": MANIFEST}

    source = create_source(config, limits)

    assert isinstance(source, ManifestDeclarativeSource)
    assert source._constructor._limit_pages_fetched_per_slice == limits.max_pages_per_slice
    assert source._constructor._limit_slices_fetched == limits.max_slices
    assert source.streams(config={})[0].retriever.requester.max_retries == 0


def request_log_message(request: dict) -> AirbyteMessage:
    return AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message=f"request:{json.dumps(request)}"))


def response_log_message(response: dict) -> AirbyteMessage:
    return AirbyteMessage(type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message=f"response:{json.dumps(response)}"))


def _create_request():
    url = "https://example.com/api"
    headers = {"Content-Type": "application/json"}
    return requests.Request("POST", url, headers=headers, json={"key": "value"}).prepare()


def _create_response(body, request):
    response = requests.Response()
    response.status_code = 200
    response._content = bytes(json.dumps(body), "utf-8")
    response.headers["Content-Type"] = "application/json"
    response.request = request
    return response


def _create_page_response(response_body):
    request = _create_request()
    return _create_response(response_body, request)


@patch.object(
    requests.Session,
    "send",
    side_effect=(
        _create_page_response({"result": [{"id": 0}, {"id": 1}], "_metadata": {"next": "next"}}),
        _create_page_response({"result": [{"id": 2}], "_metadata": {"next": "next"}}),
    )
    * 10,
)
def test_read_source(mock_http_stream):
    """
    This test sort of acts as an integration test for the connector builder.

    Each slice has two pages
    The first page has two records
    The second page one record

    The response._metadata.next field in the first page tells the paginator to fetch the next page.
    """
    max_records = 100
    max_pages_per_slice = 2
    max_slices = 3
    limits = TestReadLimits(max_records, max_pages_per_slice, max_slices)

    catalog = ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(name=_stream_name, json_schema={}, supported_sync_modes=[SyncMode.full_refresh]),
                sync_mode=SyncMode.full_refresh,
                destination_sync_mode=DestinationSyncMode.append,
            )
        ]
    )

    config = {"__injected_declarative_manifest": MANIFEST}

    source = create_source(config, limits)

    output_data = read_stream(source, config, catalog, limits).record.data
    slices = output_data["slices"]

    assert len(slices) == max_slices
    for s in slices:
        pages = s["pages"]
        assert len(pages) == max_pages_per_slice

        first_page, second_page = pages[0], pages[1]
        assert len(first_page["records"]) == _page_size
        assert len(second_page["records"]) == 1

    streams = source.streams(config)
    for s in streams:
        assert isinstance(s.retriever, SimpleRetrieverTestReadDecorator)


@patch.object(
    requests.Session,
    "send",
    side_effect=(
        _create_page_response({"result": [{"id": 0}, {"id": 1}], "_metadata": {"next": "next"}}),
        _create_page_response({"result": [{"id": 2}], "_metadata": {"next": "next"}}),
    ),
)
def test_read_source_single_page_single_slice(mock_http_stream):
    max_records = 100
    max_pages_per_slice = 1
    max_slices = 1
    limits = TestReadLimits(max_records, max_pages_per_slice, max_slices)

    catalog = ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(name=_stream_name, json_schema={}, supported_sync_modes=[SyncMode.full_refresh]),
                sync_mode=SyncMode.full_refresh,
                destination_sync_mode=DestinationSyncMode.append,
            )
        ]
    )

    config = {"__injected_declarative_manifest": MANIFEST}

    source = create_source(config, limits)

    output_data = read_stream(source, config, catalog, limits).record.data
    slices = output_data["slices"]

    assert len(slices) == max_slices
    for s in slices:
        pages = s["pages"]
        assert len(pages) == max_pages_per_slice

        first_page = pages[0]
        assert len(first_page["records"]) == _page_size

    streams = source.streams(config)
    for s in streams:
        assert isinstance(s.retriever, SimpleRetrieverTestReadDecorator)


@pytest.mark.parametrize(
    "deployment_mode, url_base, expected_error",
    [
        pytest.param("CLOUD", "https://airbyte.com/api/v1/characters", None, id="test_cloud_read_with_public_endpoint"),
        pytest.param("CLOUD", "https://10.0.27.27", "AirbyteTracedException", id="test_cloud_read_with_private_endpoint"),
        pytest.param("CLOUD", "https://localhost:80/api/v1/cast", "AirbyteTracedException", id="test_cloud_read_with_localhost"),
        pytest.param("CLOUD", "http://unsecured.protocol/api/v1", "InvalidSchema", id="test_cloud_read_with_unsecured_endpoint"),
        pytest.param("CLOUD", "https://domainwithoutextension", "Invalid URL", id="test_cloud_read_with_invalid_url_endpoint"),
        pytest.param("OSS", "https://airbyte.com/api/v1/", None, id="test_oss_read_with_public_endpoint"),
        pytest.param("OSS", "https://10.0.27.27/api/v1/", None, id="test_oss_read_with_private_endpoint"),
    ],
)
@patch.object(requests.Session, "send", _mocked_send)
def test_handle_read_external_requests(deployment_mode, url_base, expected_error):
    """
    This test acts like an integration test for the connector builder when it receives Test Read requests.

    The scenario being tested is whether requests should be denied if they are done on an unsecure channel or are made to internal
    endpoints when running on Cloud or OSS deployments
    """

    limits = TestReadLimits(max_records=100, max_pages_per_slice=1, max_slices=1)

    catalog = ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(name=_stream_name, json_schema={}, supported_sync_modes=[SyncMode.full_refresh]),
                sync_mode=SyncMode.full_refresh,
                destination_sync_mode=DestinationSyncMode.append,
            )
        ]
    )

    test_manifest = MANIFEST
    test_manifest["streams"][0]["$parameters"]["url_base"] = url_base
    config = {"__injected_declarative_manifest": test_manifest}

    source = create_source(config, limits)

    with mock.patch.dict(os.environ, {"DEPLOYMENT_MODE": deployment_mode}, clear=False):
        output_data = read_stream(source, config, catalog, limits).record.data
        if expected_error:
            assert len(output_data["logs"]) > 0, "Expected at least one log message with the expected error"
            error_message = output_data["logs"][0]
            assert error_message["level"] == "ERROR"
            assert expected_error in error_message["message"]
        else:
            page_records = output_data["slices"][0]["pages"][0]
            assert len(page_records) == len(MOCK_RESPONSE["result"])


@pytest.mark.parametrize(
    "deployment_mode, token_url, expected_error",
    [
        pytest.param("CLOUD", "https://airbyte.com/tokens/bearer", None, id="test_cloud_read_with_public_endpoint"),
        pytest.param("CLOUD", "https://10.0.27.27/tokens/bearer", "AirbyteTracedException", id="test_cloud_read_with_private_endpoint"),
        pytest.param("CLOUD", "http://unsecured.protocol/tokens/bearer", "InvalidSchema", id="test_cloud_read_with_unsecured_endpoint"),
        pytest.param("CLOUD", "https://domainwithoutextension", "Invalid URL", id="test_cloud_read_with_invalid_url_endpoint"),
        pytest.param("OSS", "https://airbyte.com/tokens/bearer", None, id="test_oss_read_with_public_endpoint"),
        pytest.param("OSS", "https://10.0.27.27/tokens/bearer", None, id="test_oss_read_with_private_endpoint"),
    ],
)
@patch.object(requests.Session, "send", _mocked_send)
def test_handle_read_external_oauth_request(deployment_mode, token_url, expected_error):
    """
    This test acts like an integration test for the connector builder when it receives Test Read requests.

    The scenario being tested is whether requests should be denied if they are done on an unsecure channel or are made to internal
    endpoints when running on Cloud or OSS deployments
    """

    limits = TestReadLimits(max_records=100, max_pages_per_slice=1, max_slices=1)

    catalog = ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(name=_stream_name, json_schema={}, supported_sync_modes=[SyncMode.full_refresh]),
                sync_mode=SyncMode.full_refresh,
                destination_sync_mode=DestinationSyncMode.append,
            )
        ]
    )

    oauth_authenticator_config: dict[str, str] = {
        "type": "OAuthAuthenticator",
        "token_refresh_endpoint": token_url,
        "client_id": "greta",
        "client_secret": "teo",
        "refresh_token": "john",
    }

    test_manifest = MANIFEST
    test_manifest["definitions"]["retriever"]["requester"]["authenticator"] = oauth_authenticator_config
    config = {"__injected_declarative_manifest": test_manifest}

    source = create_source(config, limits)

    with mock.patch.dict(os.environ, {"DEPLOYMENT_MODE": deployment_mode}, clear=False):
        output_data = read_stream(source, config, catalog, limits).record.data
        if expected_error:
            assert len(output_data["logs"]) > 0, "Expected at least one log message with the expected error"
            error_message = output_data["logs"][0]
            assert error_message["level"] == "ERROR"
            assert expected_error in error_message["message"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from airbyte_cdk.sources.streams.concurrent.cursor import CursorField
from unit_tests.sources.file_based.scenarios.scenario_builder import IncrementalScenarioConfig, TestScenarioBuilder
from unit_tests.sources.streams.concurrent.scenarios.stream_facade_builder import StreamFacadeSourceBuilder
from unit_tests.sources.streams.concurrent.scenarios.utils import MockStream

_stream1 = MockStream(
    [
        (None, [{"id": "1"}, {"id": "2"}]),
    ],
    "stream1",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
)

_stream_raising_exception = MockStream(
    [
        (None, [{"id": "1"}, ValueError("test exception")]),
    ],
    "stream1",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
)

_stream_with_primary_key = MockStream(
    [
        (None, [{"id": "1"}, {"id": "2"}]),
    ],
    "stream1",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
    primary_key="id",
)

_stream2 = MockStream(
    [
        (None, [{"id": "A"}, {"id": "B"}]),
    ],
    "stream2",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
)

_stream_with_single_slice = MockStream(
    [
        ({"slice_key": "s1"}, [{"id": "1"}, {"id": "2"}]),
    ],
    "stream1",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
)

_stream_with_multiple_slices = MockStream(
    [
        ({"slice_key": "s1"}, [{"id": "1"}, {"id": "2"}]),
        ({"slice_key": "s2"}, [{"id": "3"}, {"id": "4"}]),
    ],
    "stream1",
    json_schema={
        "type": "object",
        "properties": {
            "id": {"type": ["null", "string"]},
        },
    },
)

test_stream_facade_single_stream = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_single_stream")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream1]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .set_expected_logs(
        {
            "read": [
                {"level": "INFO", "message": "Starting syncing"},
                {"level": "INFO", "message": "Marking stream stream1 as STARTED"},
                {"level": "INFO", "message": "Syncing stream: stream1"},
                {"level": "INFO", "message": "Marking stream stream1 as RUNNING"},
                {"level": "INFO", "message": "Read 2 records from stream1 stream"},
                {"level": "INFO", "message": "Marking stream stream1 as STOPPED"},
                {"level": "INFO", "message": "Finished syncing stream1"},
                {"level": "INFO", "message": "Finished syncing"},
            ]
        }
    )
    .set_log_levels({"ERROR", "WARN", "WARNING", "INFO", "DEBUG"})
    .build()
)

test_stream_facade_raises_exception = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_raises_exception")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream_raising_exception]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .set_expected_read_error(ValueError, "test exception")
    .build()
)

test_stream_facade_single_stream_with_primary_key = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_stream_with_primary_key")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream1]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .build()
)

test_stream_facade_multiple_streams = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_multiple_streams")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream1, _stream2]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
            {"data": {"id": "A"}, "stream": "stream2"},
            {"data": {"id": "B"}, "stream": "stream2"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                },
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream2",
                    "supported_sync_modes": ["full_refresh"],
                },
            ]
        }
    )
    .build()
)

test_stream_facade_single_stream_with_single_slice = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_single_stream_with_single_slice")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream1]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .build()
)

test_stream_facade_single_stream_with_multiple_slices = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_single_stream_with_multiple_slice")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream_with_multiple_slices]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
            {"data": {"id": "3"}, "stream": "stream1"},
            {"data": {"id": "4"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .build()
)

test_stream_facade_single_stream_with_multiple_slices_with_concurrency_level_two = (
    TestScenarioBuilder()
    .set_name("test_stream_facade_single_stream_with_multiple_slice_with_concurrency_level_two")
    .set_config({})
    .set_source_builder(StreamFacadeSourceBuilder().set_streams([_stream_with_multiple_slices]))
    .set_expected_records(
        [
            {"data": {"id": "1"}, "stream": "stream1"},
            {"data": {"id": "2"}, "stream": "stream1"},
            {"data": {"id": "3"}, "stream": "stream1"},
            {"data": {"id": "4"}, "stream": "stream1"},
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                    "name": "stream1",
                    "supported_sync_modes": ["full_refresh"],
                }
            ]
        }
    )
    .build()
)


test_incremental_stream_with_slice_boundaries = (
    TestScenarioBuilder()
    .set_name("test_incremental_stream_with_slice_boundaries")
    .set_config({})
    .set_source_builder(
        StreamFacadeSourceBuilder()
        .set_streams(
            [
                MockStream(
                    [
                        ({"from": 0, "to": 1}, [{"id": "1", "cursor_field": 0}, {"id": "2", "cursor_field": 1}]),
                        ({"from": 1, "to": 2}, [{"id": "3", "cursor_field": 2}, {"id": "4", "cursor_field": 3}]),
                    ],
                    "stream1",
                    json_schema={
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                )
            ]
        )
        .set_incremental(CursorField("cursor_field"), ("from", "to"))
    )
    .set_expected_records(
        [
            {"data": {"id": "1", "cursor_field": 0}, "stream": "stream1"},
            {"data": {"id": "2", "cursor_field": 1}, "stream": "stream1"},
            {"stream1": {'cursor_field': 1}},
            {"data": {"id": "3", "cursor_field": 2}, "stream": "stream1"},
            {"data": {"id": "4", "cursor_field": 3}, "stream": "stream1"},
            {"stream1": {"cursor_field": 2}},
            {"stream1": {"cursor_field": 2}},  # see Cursor.ensure_at_least_one_state_emitted
        ]
    )
    .set_log_levels({"ERROR", "WARN", "WARNING", "INFO", "DEBUG"})
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
    .build()
)


_NO_SLICE_BOUNDARIES = None
test_incremental_stream_without_slice_boundaries = (
    TestScenarioBuilder()
    .set_name("test_incremental_stream_without_slice_boundaries")
    .set_config({})
    .set_source_builder(
        StreamFacadeSourceBuilder()
        .set_streams(
            [
                MockStream(
                    [
                        (None, [{"id": "1", "cursor_field": 0}, {"id": "2", "cursor_field": 3}]),
                    ],
                    "stream1",
                    json_schema={
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                )
            ]
        )
        .set_incremental(CursorField("cursor_field"), _NO_SLICE_BOUNDARIES)
    )
    .set_expected_records(
        [
            {"data": {"id": "1", "cursor_field": 0}, "stream": "stream1"},
            {"data": {"id": "2", "cursor_field": 3}, "stream": "stream1"},
            {"stream1": {"cursor_field": 3}},
            {"stream1": {"cursor_field": 3}},  # see Cursor.ensure_at_least_one_state_emitted
        ]
    )
    .set_log_levels({"ERROR", "WARN", "WARNING", "INFO", "DEBUG"})
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
    .build()
)

test_incremental_stream_with_many_slices_but_without_slice_boundaries = (
    TestScenarioBuilder()
    .set_name("test_incremental_stream_with_many_slices_byt_without_slice_boundaries")
    .set_config({})
    .set_source_builder(
        StreamFacadeSourceBuilder()
        .set_streams(
            [
                MockStream(
                    [
                        ({"parent_id": 1}, [{"id": "1", "cursor_field": 0}]),
                        ({"parent_id": 309}, [{"id": "3", "cursor_field": 0}]),
                    ],
                    "stream1",
                    json_schema={
                        "type": "object",
                        "properties": {
                            "id": {"type": ["null", "string"]},
                        },
                    },
                )
            ]
        )
        .set_incremental(CursorField("cursor_field"), _NO_SLICE_BOUNDARIES)
    )
    .set_expected_read_error(ValueError, "test exception")
    .set_log_levels({"ERROR", "WARN", "WARNING", "INFO", "DEBUG"})
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
    .build()
)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import unittest
from unittest.mock import Mock

from airbyte_cdk.sources.file_based.discovery_policy.default_discovery_policy import DefaultDiscoveryPolicy
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser


class DefaultDiscoveryPolicyTest(unittest.TestCase):
    def setUp(self) -> None:
        self._policy = DefaultDiscoveryPolicy()

        self._parser = Mock(spec=FileTypeParser)
        self._parser.parser_max_n_files_for_schema_inference = None

    def test_hardcoded_schema_inference_file_limit_is_returned(self) -> None:
        """
        If the parser is not providing a limit, then we should use the hardcoded limit
        """
        assert self._policy.get_max_n_files_for_schema_inference(self._parser) == 10

    def test_parser_limit_is_respected(self) -> None:
        """
        If the parser is providing a limit, then we should use that limit
        """
        self._parser.parser_max_n_files_for_schema_inference = 1

        assert self._policy.get_max_n_files_for_schema_inference(self._parser) == 1



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import os
import tempfile
import time
from datetime import datetime, timedelta
from typing import Iterable, Mapping

import pytest
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.streams.call_rate import (
    APIBudget,
    CallRateLimitHit,
    FixedWindowCallRatePolicy,
    HttpRequestMatcher,
    MovingWindowCallRatePolicy,
    Rate,
    UnlimitedCallRatePolicy,
)
from airbyte_cdk.sources.streams.http import HttpStream
from airbyte_cdk.sources.streams.http.requests_native_auth import TokenAuthenticator
from airbyte_cdk.utils.constants import ENV_REQUEST_CACHE_PATH
from requests import Request


class StubDummyHttpStream(HttpStream):
    url_base = "https://test_base_url.com"
    primary_key = "some_key"

    def next_page_token(self, *args, **kwargs):
        return True  # endless pages

    def path(self, **kwargs) -> str:
        return ""

    def parse_response(self, *args, **kwargs) -> Iterable[Mapping]:
        yield {"data": "some_data"}


class StubDummyCacheHttpStream(StubDummyHttpStream):
    use_cache = True


@pytest.fixture(name="enable_cache")
def enable_cache_fixture():
    prev_cache_path = os.environ.get(ENV_REQUEST_CACHE_PATH)
    with tempfile.TemporaryDirectory() as temp_dir:
        os.environ[ENV_REQUEST_CACHE_PATH] = temp_dir
        yield

    if prev_cache_path is not None:
        os.environ[ENV_REQUEST_CACHE_PATH] = prev_cache_path


class TestHttpRequestMatcher:
    try_all_types_of_requests = pytest.mark.parametrize(
        "request_factory",
        [Request, lambda *args, **kwargs: Request(*args, **kwargs).prepare()],
    )

    @try_all_types_of_requests
    def test_url(self, request_factory):
        matcher = HttpRequestMatcher(url="http://some_url/")
        assert not matcher(request_factory(url="http://some_wrong_url"))
        assert matcher(request_factory(url="http://some_url"))

    @try_all_types_of_requests
    def test_method(self, request_factory):
        matcher = HttpRequestMatcher(method="GET")
        assert not matcher(request_factory(url="http://some_url"))
        assert not matcher(request_factory(url="http://some_url", method="POST"))
        assert matcher(request_factory(url="http://some_url", method="GET"))

    @try_all_types_of_requests
    def test_params(self, request_factory):
        matcher = HttpRequestMatcher(params={"param1": 10, "param2": 15})
        assert not matcher(request_factory(url="http://some_url/"))
        assert not matcher(request_factory(url="http://some_url/", params={"param1": 10, "param3": 100}))
        assert not matcher(request_factory(url="http://some_url/", params={"param1": 10, "param2": 10}))
        assert matcher(request_factory(url="http://some_url/", params={"param1": 10, "param2": 15, "param3": 100}))

    @try_all_types_of_requests
    def test_header(self, request_factory):
        matcher = HttpRequestMatcher(headers={"header1": 10, "header2": 15})
        assert not matcher(request_factory(url="http://some_url"))
        assert not matcher(request_factory(url="http://some_url", headers={"header1": "10", "header3": "100"}))
        assert not matcher(request_factory(url="http://some_url", headers={"header1": "10", "header2": "10"}))
        assert matcher(request_factory(url="http://some_url", headers={"header1": "10", "header2": "15", "header3": "100"}))

    @try_all_types_of_requests
    def test_combination(self, request_factory):
        matcher = HttpRequestMatcher(method="GET", url="http://some_url/", headers={"header1": 10}, params={"param2": "test"})
        assert matcher(request_factory(method="GET", url="http://some_url", headers={"header1": "10"}, params={"param2": "test"}))
        assert not matcher(request_factory(method="GET", url="http://some_url", headers={"header1": "10"}))
        assert not matcher(request_factory(method="GET", url="http://some_url"))
        assert not matcher(request_factory(url="http://some_url"))


def test_http_request_matching(mocker):
    """Test policy lookup based on matchers."""
    users_policy = mocker.Mock(spec=MovingWindowCallRatePolicy)
    groups_policy = mocker.Mock(spec=MovingWindowCallRatePolicy)
    root_policy = mocker.Mock(spec=MovingWindowCallRatePolicy)

    users_policy.matches.side_effect = HttpRequestMatcher(url="http://domain/api/users", method="GET")
    groups_policy.matches.side_effect = HttpRequestMatcher(url="http://domain/api/groups", method="POST")
    root_policy.matches.side_effect = HttpRequestMatcher(method="GET")
    api_budget = APIBudget(
        policies=[
            users_policy,
            groups_policy,
            root_policy,
        ]
    )

    api_budget.acquire_call(Request("POST", url="http://domain/unmatched_endpoint"), block=False), "unrestricted call"
    users_policy.try_acquire.assert_not_called()
    groups_policy.try_acquire.assert_not_called()
    root_policy.try_acquire.assert_not_called()

    users_request = Request("GET", url="http://domain/api/users")
    api_budget.acquire_call(users_request, block=False), "first call, first matcher"
    users_policy.try_acquire.assert_called_once_with(users_request, weight=1)
    groups_policy.try_acquire.assert_not_called()
    root_policy.try_acquire.assert_not_called()

    api_budget.acquire_call(Request("GET", url="http://domain/api/users"), block=False), "second call, first matcher"
    assert users_policy.try_acquire.call_count == 2
    groups_policy.try_acquire.assert_not_called()
    root_policy.try_acquire.assert_not_called()

    group_request = Request("POST", url="http://domain/api/groups")
    api_budget.acquire_call(group_request, block=False), "first call, second matcher"
    assert users_policy.try_acquire.call_count == 2
    groups_policy.try_acquire.assert_called_once_with(group_request, weight=1)
    root_policy.try_acquire.assert_not_called()

    api_budget.acquire_call(Request("POST", url="http://domain/api/groups"), block=False), "second call, second matcher"
    assert users_policy.try_acquire.call_count == 2
    assert groups_policy.try_acquire.call_count == 2
    root_policy.try_acquire.assert_not_called()

    any_get_request = Request("GET", url="http://domain/api/")
    api_budget.acquire_call(any_get_request, block=False), "first call, third matcher"
    assert users_policy.try_acquire.call_count == 2
    assert groups_policy.try_acquire.call_count == 2
    root_policy.try_acquire.assert_called_once_with(any_get_request, weight=1)


class TestUnlimitedCallRatePolicy:
    def test_try_acquire(self, mocker):
        policy = UnlimitedCallRatePolicy(matchers=[])
        assert policy.matches(mocker.Mock()), "should match anything"
        policy.try_acquire(mocker.Mock(), weight=1)
        policy.try_acquire(mocker.Mock(), weight=10)

    def test_update(self):
        policy = UnlimitedCallRatePolicy(matchers=[])
        policy.update(available_calls=10, call_reset_ts=datetime.now())
        policy.update(available_calls=None, call_reset_ts=datetime.now())
        policy.update(available_calls=10, call_reset_ts=None)


class TestFixedWindowCallRatePolicy:
    def test_limit_rate(self, mocker):
        policy = FixedWindowCallRatePolicy(matchers=[], next_reset_ts=datetime.now(), period=timedelta(hours=1), call_limit=100)
        policy.try_acquire(mocker.Mock(), weight=1)
        policy.try_acquire(mocker.Mock(), weight=20)
        with pytest.raises(ValueError, match="Weight can not exceed the call limit"):
            policy.try_acquire(mocker.Mock(), weight=101)

        with pytest.raises(CallRateLimitHit) as exc:
            policy.try_acquire(mocker.Mock(), weight=100 - 20 - 1 + 1)

        assert exc.value.time_to_wait
        assert exc.value.weight == 100 - 20 - 1 + 1
        assert exc.value.item

    def test_update_available_calls(self, mocker):
        policy = FixedWindowCallRatePolicy(matchers=[], next_reset_ts=datetime.now(), period=timedelta(hours=1), call_limit=100)
        # update to decrease number of calls available
        policy.update(available_calls=2, call_reset_ts=None)
        # hit the limit with weight=3
        with pytest.raises(CallRateLimitHit):
            policy.try_acquire(mocker.Mock(), weight=3)
        # ok with less weight=1
        policy.try_acquire(mocker.Mock(), weight=1)

        # update to increase number of calls available, ignored
        policy.update(available_calls=20, call_reset_ts=None)
        # so we still hit the limit with weight=3
        with pytest.raises(CallRateLimitHit):
            policy.try_acquire(mocker.Mock(), weight=3)


class TestMovingWindowCallRatePolicy:
    def test_no_rates(self):
        """should raise a ValueError when no rates provided"""
        with pytest.raises(ValueError, match="The list of rates can not be empty"):
            MovingWindowCallRatePolicy(rates=[], matchers=[])

    def test_limit_rate(self):
        """try_acquire must respect configured call rate and throw CallRateLimitHit when hit the limit."""
        policy = MovingWindowCallRatePolicy(rates=[Rate(10, timedelta(minutes=1))], matchers=[])

        for i in range(10):
            policy.try_acquire("call", weight=1), f"{i + 1} call"

        with pytest.raises(CallRateLimitHit) as excinfo1:
            policy.try_acquire("call", weight=1), "call over limit"
        assert excinfo1.value.time_to_wait.total_seconds() == pytest.approx(60, 0.1)

        time.sleep(0.5)

        with pytest.raises(CallRateLimitHit) as excinfo2:
            policy.try_acquire("call", weight=1), "call over limit"
        assert excinfo2.value.time_to_wait < excinfo1.value.time_to_wait, "time to wait must decrease over time"

    def test_limit_rate_support_custom_weight(self):
        """try_acquire must take into account provided weight and throw CallRateLimitHit when hit the limit."""
        policy = MovingWindowCallRatePolicy(rates=[Rate(10, timedelta(minutes=1))], matchers=[])

        policy.try_acquire("call", weight=2), "1st call with weight of 2"
        with pytest.raises(CallRateLimitHit) as excinfo:
            policy.try_acquire("call", weight=9), "2nd call, over limit since 2 + 9 = 11 > 10"
        assert excinfo.value.time_to_wait.total_seconds() == pytest.approx(60, 0.1), "should wait 1 minute before next call"

    def test_multiple_limit_rates(self):
        """try_acquire must take into all call rates and apply stricter."""
        policy = MovingWindowCallRatePolicy(
            matchers=[],
            rates=[
                Rate(10, timedelta(minutes=10)),
                Rate(3, timedelta(seconds=10)),
                Rate(2, timedelta(hours=1)),
            ],
        )

        policy.try_acquire("call", weight=2), "1 call"

        with pytest.raises(CallRateLimitHit) as excinfo:
            policy.try_acquire("call", weight=1), "1 call"

        assert excinfo.value.time_to_wait.total_seconds() == pytest.approx(3600, 0.1)
        assert str(excinfo.value) == "Bucket for item=call with Rate limit=2/1.0h is already full"


class TestHttpStreamIntegration:
    def test_without_cache(self, mocker, requests_mock):
        """Test that HttpStream will use call budget when provided"""
        requests_mock.get(f"{StubDummyHttpStream.url_base}/", json={"data": "test"})

        mocker.patch.object(MovingWindowCallRatePolicy, "try_acquire")

        api_budget = APIBudget(
            policies=[
                MovingWindowCallRatePolicy(
                    matchers=[HttpRequestMatcher(url=f"{StubDummyHttpStream.url_base}/", method="GET")],
                    rates=[
                        Rate(2, timedelta(minutes=1)),
                    ],
                ),
            ]
        )

        stream = StubDummyHttpStream(api_budget=api_budget, authenticator=TokenAuthenticator(token="ABCD"))
        records = stream.read_records(SyncMode.full_refresh)
        for i in range(10):
            assert next(records) == {"data": "some_data"}

        assert MovingWindowCallRatePolicy.try_acquire.call_count == 10

    @pytest.mark.usefixtures("enable_cache")
    def test_with_cache(self, mocker, requests_mock):
        """Test that HttpStream will use call budget when provided and not cached"""
        requests_mock.get(f"{StubDummyHttpStream.url_base}/", json={"data": "test"})

        mocker.patch.object(MovingWindowCallRatePolicy, "try_acquire")

        api_budget = APIBudget(
            policies=[
                MovingWindowCallRatePolicy(
                    matchers=[
                        HttpRequestMatcher(url=f"{StubDummyHttpStream.url_base}/", method="GET"),
                    ],
                    rates=[
                        Rate(2, timedelta(minutes=1)),
                    ],
                )
            ]
        )

        stream = StubDummyCacheHttpStream(api_budget=api_budget)
        records = stream.read_records(SyncMode.full_refresh)

        for i in range(10):
            assert next(records) == {"data": "some_data"}

        assert MovingWindowCallRatePolicy.try_acquire.call_count == 1



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union
from unittest.mock import Mock

import freezegun
from airbyte_cdk.models import (
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteStream,
    AirbyteStreamStatus,
    AirbyteStreamStatusTraceMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    DestinationSyncMode,
    StreamDescriptor,
    SyncMode,
    TraceType,
)
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.concurrent_source.concurrent_source import ConcurrentSource
from airbyte_cdk.sources.concurrent_source.concurrent_source_adapter import ConcurrentSourceAdapter
from airbyte_cdk.sources.message import InMemoryMessageRepository
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.concurrent.adapters import StreamFacade
from airbyte_cdk.sources.streams.concurrent.cursor import NoopCursor
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.utils import AirbyteTracedException
from unit_tests.sources.streams.concurrent.scenarios.thread_based_concurrent_stream_source_builder import NeverLogSliceLogger


class _MockStream(Stream):
    def __init__(self, slice_to_records: Mapping[str, List[Mapping[str, Any]]], name: str):
        self._slice_to_records = slice_to_records
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        return None

    def stream_slices(
        self, *, sync_mode: SyncMode, cursor_field: Optional[List[str]] = None, stream_state: Optional[Mapping[str, Any]] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        for partition in self._slice_to_records.keys():
            yield {"partition": partition}

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: Optional[List[str]] = None,
        stream_slice: Optional[Mapping[str, Any]] = None,
        stream_state: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[StreamData]:
        for record_or_exception in self._slice_to_records[stream_slice["partition"]]:
            if isinstance(record_or_exception, Exception):
                raise record_or_exception
            else:
                yield record_or_exception

    def get_json_schema(self) -> Mapping[str, Any]:
        return {}


class _MockSource(AbstractSource):
    message_repository = InMemoryMessageRepository()

    def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Optional[Any]]:
        pass

    def set_streams(self, streams):
        self._streams = streams

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        return self._streams


class _MockConcurrentSource(ConcurrentSourceAdapter):
    message_repository = InMemoryMessageRepository()

    def __init__(self, logger):
        concurrent_source = ConcurrentSource.create(1, 1, logger, NeverLogSliceLogger(), self.message_repository)
        super().__init__(concurrent_source)

    def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Optional[Any]]:
        pass

    def set_streams(self, streams):
        self._streams = streams

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        return self._streams


@freezegun.freeze_time("2020-01-01T00:00:00")
def test_concurrent_source_yields_the_same_messages_as_abstract_source_when_no_exceptions_are_raised():
    records_stream_1_partition_1 = [
        {"id": 1, "partition": "1"},
        {"id": 2, "partition": "1"},
    ]
    records_stream_1_partition_2 = [
        {"id": 3, "partition": "2"},
        {"id": 4, "partition": "2"},
    ]
    records_stream_2_partition_1 = [
        {"id": 100, "partition": "A"},
        {"id": 200, "partition": "A"},
    ]
    records_stream_2_partition_2 = [
        {"id": 300, "partition": "B"},
        {"id": 400, "partition": "B"},
    ]
    stream_1_slice_to_partition = {"1": records_stream_1_partition_1, "2": records_stream_1_partition_2}
    stream_2_slice_to_partition = {"A": records_stream_2_partition_1, "B": records_stream_2_partition_2}
    state = None
    logger = _init_logger()

    source, concurrent_source = _init_sources([stream_1_slice_to_partition, stream_2_slice_to_partition], state, logger)

    config = {}
    catalog = _create_configured_catalog(source._streams)
    messages_from_abstract_source = _read_from_source(source, logger, config, catalog, state, None)
    messages_from_concurrent_source = _read_from_source(concurrent_source, logger, config, catalog, state, None)

    expected_messages = [
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.STARTED)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.RUNNING)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records_stream_1_partition_1[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records_stream_1_partition_1[1],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records_stream_1_partition_2[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records_stream_1_partition_2[1],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.COMPLETE)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream1"), status=AirbyteStreamStatus(AirbyteStreamStatus.STARTED)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream1"), status=AirbyteStreamStatus(AirbyteStreamStatus.RUNNING)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream1",
                data=records_stream_2_partition_1[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream1",
                data=records_stream_2_partition_1[1],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream1",
                data=records_stream_2_partition_2[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream1",
                data=records_stream_2_partition_2[1],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream1"), status=AirbyteStreamStatus(AirbyteStreamStatus.COMPLETE)
                ),
            ),
        ),
    ]
    _verify_messages(expected_messages, messages_from_abstract_source, messages_from_concurrent_source)


@freezegun.freeze_time("2020-01-01T00:00:00")
def test_concurrent_source_yields_the_same_messages_as_abstract_source_when_a_traced_exception_is_raised():
    records = [{"id": 1, "partition": "1"}, AirbyteTracedException()]
    stream_slice_to_partition = {"1": records}

    logger = _init_logger()
    state = None
    source, concurrent_source = _init_sources([stream_slice_to_partition], state, logger)
    config = {}
    catalog = _create_configured_catalog(source._streams)
    messages_from_abstract_source = _read_from_source(source, logger, config, catalog, state, AirbyteTracedException)
    messages_from_concurrent_source = _read_from_source(concurrent_source, logger, config, catalog, state, AirbyteTracedException)

    expected_messages = [
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.STARTED)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.RUNNING)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.INCOMPLETE)
                ),
            ),
        ),
    ]
    _verify_messages(expected_messages, messages_from_abstract_source, messages_from_concurrent_source)


@freezegun.freeze_time("2020-01-01T00:00:00")
def test_concurrent_source_yields_the_same_messages_as_abstract_source_when_an_exception_is_raised():
    records = [{"id": 1, "partition": "1"}, RuntimeError()]
    stream_slice_to_partition = {"1": records}
    logger = _init_logger()

    state = None

    source, concurrent_source = _init_sources([stream_slice_to_partition], state, logger)
    config = {}
    catalog = _create_configured_catalog(source._streams)
    messages_from_abstract_source = _read_from_source(source, logger, config, catalog, state, RuntimeError)
    messages_from_concurrent_source = _read_from_source(concurrent_source, logger, config, catalog, state, RuntimeError)

    expected_messages = [
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.STARTED)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.RUNNING)
                ),
            ),
        ),
        AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                stream="stream0",
                data=records[0],
                emitted_at=1577836800000,
            ),
        ),
        AirbyteMessage(
            type=MessageType.TRACE,
            trace=AirbyteTraceMessage(
                type=TraceType.STREAM_STATUS,
                emitted_at=1577836800000.0,
                error=None,
                estimate=None,
                stream_status=AirbyteStreamStatusTraceMessage(
                    stream_descriptor=StreamDescriptor(name="stream0"), status=AirbyteStreamStatus(AirbyteStreamStatus.INCOMPLETE)
                ),
            ),
        ),
    ]
    _verify_messages(expected_messages, messages_from_abstract_source, messages_from_concurrent_source)


def _init_logger():
    logger = Mock()
    logger.level = logging.INFO
    logger.isEnabledFor.return_value = False
    return logger


def _init_sources(stream_slice_to_partitions, state, logger):
    source = _init_source(stream_slice_to_partitions, state, logger, _MockSource())
    concurrent_source = _init_source(stream_slice_to_partitions, state, logger, _MockConcurrentSource(logger))
    return source, concurrent_source


def _init_source(stream_slice_to_partitions, state, logger, source):
    cursor = NoopCursor()
    streams = [
        StreamFacade.create_from_stream(_MockStream(stream_slices, f"stream{i}"), source, logger, state, cursor)
        for i, stream_slices in enumerate(stream_slice_to_partitions)
    ]
    source.set_streams(streams)
    return source


def _create_configured_catalog(streams):
    return ConfiguredAirbyteCatalog(
        streams=[
            ConfiguredAirbyteStream(
                stream=AirbyteStream(name=s.name, json_schema={}, supported_sync_modes=[SyncMode.full_refresh]),
                sync_mode=SyncMode.full_refresh,
                cursor_field=None,
                destination_sync_mode=DestinationSyncMode.overwrite,
            )
            for s in streams
        ]
    )


def _read_from_source(source, logger, config, catalog, state, expected_exception):
    messages = []
    try:
        for m in source.read(logger, config, catalog, state):
            messages.append(m)
    except Exception as e:
        if expected_exception:
            assert isinstance(e, expected_exception)
    return messages


def _verify_messages(expected_messages, messages_from_abstract_source, messages_from_concurrent_source):
    assert _compare(expected_messages, messages_from_concurrent_source)


def _compare(s, t):
    # Use a compare method that does not require ordering or hashing the elements
    # We can't rely on the ordering because of the multithreading
    # AirbyteMessage does not implement __eq__ and __hash__
    t = list(t)
    try:
        for elem in s:
            t.remove(elem)
    except ValueError:
        print(f"ValueError: {elem}")
        return False
    return not t
